#
# Stanislav Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact s.georgiev@softel.bg.
#
# 2025
#

import time
import os
import shutil
import cv2
import numpy as np

from sciveo.tools.logger import *
from sciveo.tools.daemon import DaemonBase
from sciveo.media.capture.readers import VideoReaderFFMPEG
from sciveo.tools.queue import TouchedFilePathQueue
from sciveo.tools.simple_counter import SimpleStat


class MotionDetectorGrid:
  def __init__(self, grid):
    self.grid = np.array(grid)
    self.grid_mask = None
    self.zero = False

  def init_grid(self, frame):
    self.grid_mask = np.zeros_like(frame, dtype=np.float32)

    for i in range(0, self.grid_mask.shape[0]):
      for j in range(0, self.grid_mask.shape[1]):
        gi = int(self.grid.shape[0] * i / self.grid_mask.shape[0])
        gj = int(self.grid.shape[1] * j / self.grid_mask.shape[1])

        self.grid_mask[i][j] = self.grid[gi][gj]

    self.zero = self.grid_mask.sum() <= 0.0

  def apply(self, frame):
    if self.grid_mask is None:
      self.init_grid(frame)
    result = (frame * self.grid_mask).astype(np.uint8)
    return result


class VideoMotionDetector:
  def __init__(self, source_id, configuration):
    super().__init__()
    self.source_id = source_id
    self.configuration = configuration
    self.count_motion_hold = 0
    self.detection_grid = MotionDetectorGrid(self.configuration["grid"])

    self.init()

  def init(self):
    self.fgbg = cv2.createBackgroundSubtractorMOG2(history=50, varThreshold=20, detectShadows=False)

  def apply(self, frame):
    md_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
    md_frame = cv2.blur(src=md_frame, ksize=(10,10))
    md_frame = self.fgbg.apply(md_frame)
    md_frame = self.detection_grid.apply(md_frame)
    return md_frame

  def check_motion(self, fgmask):
    k_motion_level = 0.5 + 0.5 * min(self.configuration["threshold_motion_hold"], self.count_motion_hold) / self.configuration["threshold_motion_hold"]
    current_threshold = self.configuration["threshold_motion"] * k_motion_level
    motion_level = np.mean(fgmask)

    self.stat.set(motion_level)

    if motion_level > current_threshold:
      self.count_motion_hold += 1
      if self.count_motion_hold >= self.configuration["threshold_motion_hold"]:
        debug(self.source_id, "motion detected", motion_level, "threshold", current_threshold)
        return True
    else:
      self.count_motion_hold = 0
    return False

  def __call__(self, list_frames):
    self.count_motion_hold = 0
    self.stat = SimpleStat()

    if len(list_frames) > 0:
      for i in range(10):
        fgmask = self.apply(list_frames[0])

    for frame in list_frames:
      try:
        if self.detection_grid.zero:
          continue

        fgmask = self.apply(frame)

        if self.check_motion(fgmask):
          return True
      except Exception as e:
        exception(e)
        time.sleep(3)

    return False


class MotionDetectorDaemon(DaemonBase):
  def __init__(self, configuration):
    super().__init__()
    self.configuration = configuration
    self.queue = TouchedFilePathQueue(self.configuration["src"], period=3, touched_timeout=3)
    self.motion_detectors = {}

  def on_motion(self, source_id, file_name, file_path, configuration):
    debug("MOTION", source_id, file_name, configuration)
    if "dst" in self.configuration:
      file_name_split = file_name.split("___")
      if len(file_name_split) >= 3:
        cam_id = file_name_split[0]
        video_date = file_name_split[1]
        video_file_name = file_name_split[2]
        video_base_path = os.path.join(self.configuration["dst"], cam_id, video_date)
        dst_file_path = os.path.join(video_base_path, video_file_name)
        os.makedirs(video_base_path, exist_ok=True)
        shutil.copy(file_path, dst_file_path)
        debug("CP", file_path, "=>", dst_file_path)

  def run(self):
    while(self.is_running):
      try:
        file_name, file_path = self.queue.pop()
        debug("pop", file_name, file_path)

        try:
          file_name_split = file_name.split("___")
          if len(file_name_split) >= 3:
            source_id = file_name_split[0]
            if source_id in self.configuration:
              configuration = self.configuration.get(source_id, { "threshold_motion": 5, "threshold_motion_hold": 5, "grid": [[0.0]] })
              md = self.motion_detectors.get(source_id, VideoMotionDetector(source_id, configuration))
              motion = md(VideoReaderFFMPEG.read(file_path, resolution=360, RGB=False, gpu_id=-1, framestep=5))
              debug("STAT", source_id, md.stat)
              if motion:
                self.on_motion(source_id, file_name, file_path, configuration)
        except Exception as e:
          exception(e)
        os.remove(file_path)
        debug("RM", file_path)
      except Exception as e:
        exception(e)
        time.sleep(5)



#
# Stanislav Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact s.georgiev@softel.bg.
#
# 2025
#

import time
import threading
import time
from pymodbus.server import StartTcpServer
from pymodbus.datastore import ModbusServerContext, ModbusSparseDataBlock
from threading import Thread
from sciveo.tools.logger import *
from sciveo.tools.daemon import DaemonBase


# Example read-only registers
# INITIAL_REGISTERS = {
#     8044: 12500,   # Plant Active Power
#     8060: 8200,    # PV Active Power
#     8038: -1500,   # Storage Power
#     8048: 600,     # Reactive Power
#     9082: 3000,    # Dynamic Allowable Charge Power
#     9084: 3500,    # Dynamic Allowable Discharge Power
# }
INITIAL_REGISTERS = {
  8060: 8200,    # PV Active Power
}

class SimulatorEMS300(DaemonBase):
  def __init__(self, host="127.0.0.1", port=5020, period=1):
    super().__init__(num_threads=1, period=period)
    self.host = host
    self.port = port

    # Only input registers
    self.ir_block = ModbusSparseDataBlock(INITIAL_REGISTERS.copy())

    self.context = ModbusServerContext(
      { 247: {  # unit_id -> datastore dict
          'di': ModbusSparseDataBlock({}),   # discrete inputs
          'co': ModbusSparseDataBlock({}),   # coils
          'hr': ModbusSparseDataBlock({}),   # holding registers
          'ir': self.ir_block                # input registers
        }
      },
      single=False  # multiple units not used here, keep False
    )

  def loop(self):
    # Slowly vary PV Active Power
    pv = self.ir_block.getValues(8060, count=1)[0]
    pv = (pv + 50) % 15000
    self.ir_block.setValues(8060, [pv])
    info(f"Simulated PV Active Power: {pv}")

  def start_server(self):
    StartTcpServer(self.context, address=(self.host, self.port))


if __name__ == "__main__":
    sim = SimulatorEMS300(host="127.0.0.1", port=1502)
    sim.start()
    sim.start_server()
    while True:
      time.sleep(10)



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

from sciveo.tools.logger import *
from sciveo.tools.os import add_suffix_to_filename
from sciveo.media.pipelines.queues import MediaJobState
from sciveo.media.pipelines.base import BaseContentProcessor


class BaseProcessor(BaseContentProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)
    self.base_tmp_path = os.environ["WEB_MEDIA_PIPELINES_TMP_PATH"]
    debug("init", processor_config, max_progress)

  def process(self, media):
    return media

  def post_process(self):
    pass

  def init_run(self):
    pass

  def run(self, job, input):
    self.job = job
    self.job_id = job["id"]
    next_media = []
    if len(input) == 0:
      return []
    progress_per_media = self.max_progress / max(1, len(input))
    debug("run", job["id"], progress_per_media, "input", input)

    self.init_run()

    for i, media in enumerate(input):
      if self.is_processor_run(job, media):
        try:
          media = self.process(media)

          if self.is_append_processor():
            self.append_processor(media)

          if "next" not in media:
            continue

          self.resized(job, media)

          next_media += media["next"]
          media["next"] = []
        except Exception as e:
          exception(e, media)
      MediaJobState.queue().inc_progress(job["id"], progress_per_media)

    if self["output"]:
      job["output"] += next_media
    if self["append-content"]:
      job["append-content"] += next_media
    if self["update-content"]:
      job["update-content-data"] += input

    return next_media

  def resized(self, job, parent_media):
    if self.is_resizer():
      parent_media.setdefault("heights", [])
      for media in parent_media["next"]:
        parent_media["heights"].append(media["h"])
      if len(parent_media["heights"]) > 0:
        job["resized-resolutions"].append(parent_media)

  def is_resizer(self):
    return False

  def add_suffix_to_filename(self, file_path, siffix):
    if self["output"] or self["append-content"]:
      return add_suffix_to_filename(file_path, siffix)
    else:
      return file_path



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import time
import numpy as np
import cv2
from picamera2 import Picamera2

from sciveo.tools.logger import *
from sciveo.media.capture.cam import CameraDaemon
from sciveo.media.capture.gst_server import CamFactory


class RPICamera(CameraDaemon):
  def __init__(self, cam_id=0, width=640, height=480, period=0):
    self.width = width
    self.height = height
    super().__init__(cam_id=cam_id, period=period)

  def init_cam(self):
    debug(self.src, "starting...")

    self.picam2 = Picamera2()
    config = self.picam2.create_video_configuration(
      main={
        "size": (self.width, self.height),
        "format": "RGB888"
      },
      buffer_count=4
    )
    self.picam2.configure(config)
    self.picam2.start()

    time.sleep(1)
    debug(self.src, "started")

  def close(self):
    self.picam2.stop()

  def read_frame(self):
    try:
      frame = self.picam2.capture_array()
      cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
      with self.lock_frame:
        self.frame = frame
    except Exception as e:
      exception(e, "CAM frame read FAIL")


class RPICameraFactory(CamFactory):
  def init_cam(self):
    self.cam = RPICamera(cam_id=self.cam_id, width=self.width, height=self.height)



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import unittest

import sciveo


class TestRunner(unittest.TestCase):
  def dummy_loop(self):
    with sciveo.open() as E:
      self.assertTrue(E.config.configuration.keys() == self.configuration.keys())

  def test_dummy_train(self):
    self.configuration = {
        "input_window": {
            "values": [10, 20, 30, 40, 50, 100, 200]
        },
        "steps": {
            "min": 1, "max": 100
        },
        "max_epochs": {
            "min": 1, "max": 3
        },
        "patience": {
            "min": 2, "max": 3
        },
        "idx": {
            "seq": 1
        }
    }

    for sampler in ["random", "grid", "auto"]:
      sciveo.start(
        project="SCIVEO Dummy Test",
        configuration=self.configuration,
        function=self.dummy_loop,
        remote=False,
        count=10,
        sampler=sampler
      )

  def test_init_args(self):
    self.configuration = {
      "lr": (1e-5, 1e-6, 11)
    }

    sciveo.start(
        project="test_init_args",
        configuration=self.configuration,
        function=self.dummy_loop,
        remote=False,
        count=10,
        sampler="auto",
        num_random_samples=20,
        next_sample_ratio=0.99,
        min_delta_score=1e-5,
        optimizer="base",
        learning_rate=0.2,
        learning_rate_decay=0.01
      )


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import time
import math
import numpy as np

from sciveo.tools.logger import *
from sciveo.ml.images.object_detection import *


def compute_iou_xyxy(box1, box2):
  """
    box1, box2: [x1, y1, x2, y2] format.
    Returns the IoU between the two boxes.
  """
  x1_inter = max(box1[0], box2[0])
  y1_inter = max(box1[1], box2[1])
  x2_inter = min(box1[2], box2[2])
  y2_inter = min(box1[3], box2[3])

  inter_area = max(0, x2_inter - x1_inter) * max(0, y2_inter - y1_inter)
  box1_area = (box1[2] - box1[0]) * (box1[3] - box1[1])
  box2_area = (box2[2] - box2[0]) * (box2[3] - box2[1])
  union_area = box1_area + box2_area - inter_area

  if union_area == 0:
    return 0.0

  return inter_area / union_area



"""

Object Detection Evaluation

Using AP and FP for object detector evaluation.
Need to define max allowed false positives (for example 0.03).
Usually max AP has relatively high FP rate, so maxing FP is on lower AP.

"""
class EvalObjectDetection:
  def __init__(self, predictions, labels, class_names):
    self.predictions = predictions
    self.labels = labels
    self.class_names = class_names

  # Convert from yolo predictions
  def from_yolo(self):
    self.converted_predictions = []
    for predicted in self.predictions:
      boxes = predicted.boxes
      class_names = predicted.names

      """
        Converted predictions of type [{"class 1": [x1,y1,x2,y2, confidence]}, ...]
        Boxes of type xyxyn + confidence
      """
      converted_prediction = {}
      for i in range(len(boxes)):
        class_id = int(boxes.cls[i].item())
        label_class = class_names[class_id]
        if label_class not in self.class_names:
          continue
        box = boxes.xyxyn[i].tolist()
        confidence = boxes.conf[i].item()
        box_confidence = box + [confidence]

        converted_prediction.setdefault(label_class, [])
        converted_prediction[label_class].append(box_confidence)
      self.converted_predictions.append(converted_prediction)

    self._convert_labels_dict()

  """
    Labels of type [ {"class 1": [ ['x1':x1,'y1':y1,'x2':x2,'y2':y2] ... ], "class 2": [ ['x1':x1,'y1':y1,'x2':x2,'y2':y2] ... ]} ... ]
    Converted Labels of type [{"class 1": [[x1,y1,x2,y2],...], "class 2": [[], [],...]...}]
    Boxes of type xyxyn
  """
  def _convert_labels_dict(self):
    self.converted_labels = []
    for label in self.labels:
      converted_label = {}
      for class_name, class_boxes in label.items():
        converted_label[class_name] = []
        for box in class_boxes:
          converted_label[class_name].append([box['x1'], box['y1'], box['x2'], box['y2']])
      self.converted_labels.append(converted_label)

  def compute_ap(self, class_name, confidence_threshold=0.0):
    """
    Calculate the Average Precision based on IoU and confidence scores.
    Returns the AP for the current class.
    """

    count_labels = 0
    true_positives = []
    false_positives = []

    for i, prediction in enumerate(self.converted_predictions):
      current_predictions = prediction.get(class_name, [])
      current_labels = self.converted_labels[i].get(class_name, [])

      count_labels += len(current_labels)

      detected = set()

      for prediction_box in current_predictions:
        if prediction_box[4] < confidence_threshold:
          continue

        max_iou = 0
        gt_match = None

        for label_box in current_labels:
          iou = compute_iou_xyxy(prediction_box[:4], label_box)
          if iou > 0.5 and iou >= max_iou:
            max_iou = iou
            gt_match = label_box

        gt_match_key = tuple([round(x, 6) for x in gt_match]) if gt_match is not None else None

        if gt_match is not None and gt_match_key not in detected:
          true_positives.append(1)
          false_positives.append(0)
          detected.add(gt_match_key)
        else:
          true_positives.append(0)
          false_positives.append(1)

    tp_cumsum = np.cumsum(true_positives)
    fp_cumsum = np.cumsum(false_positives)

    recall = tp_cumsum / (count_labels + 1e-20)
    precision = tp_cumsum / (tp_cumsum + fp_cumsum + 1e-20)

    FP = np.sum(false_positives) / (count_labels + 1e-20)

    # Compute AP using the trapezoidal rule (integrating precision over recall)
    order = np.argsort(recall)
    try:
      ap = np.trapezoid(precision[order], recall[order])
    except Exception:
      ap = np.trapz(precision[order], recall[order])
    return ap, FP

  # TODO: Use simple gradient-based threshold optimisation instead of current grid search.
  def calc_thresholds(self, class_name):
    list_ap = []
    list_FP = []
    list_thresholds = np.linspace(0.0, 1.0, 101).tolist()
    for i, threshold in enumerate(list_thresholds):
      ap, FP = self.compute_ap(class_name, threshold)
      list_ap.append(ap)
      list_FP.append(FP)
      if i % 10 == 0:
        debug(class_name, "threshold", threshold, "ap", ap, "FP", FP)
    return list_thresholds, list_ap, list_FP

  def threshold(self, class_name, list_thresholds, list_ap, list_FP, max_fp=0.05):
    idx = list_ap.index(max(list_ap))
    if list_FP[idx] <= max_fp:
      return list_thresholds[idx]

    for i in range(len(list_thresholds)):
      if list_FP[i] <= max_fp:
        debug("Threshold", class_name, list_thresholds[i], "AP", list_ap[i], "max_fp", max_fp, "FP", list_FP[i])
        idx = i
        break

    return list_thresholds[idx]

  def thresholds(self):
    """
      Calculate best confidence thresholds for every class.
      Predictions should be non-thresholded.
      Thresholds precission is 1/100.
    """

    result = {}
    list_thresholds = {}
    list_ap = {}
    list_FP = {}

    for class_name in self.class_names:
      list_thresholds[class_name], list_ap[class_name], list_FP[class_name] = self.calc_thresholds(class_name)
      idx = list_ap[class_name].index(max(list_ap[class_name]))
      debug("Threshold", class_name, list_thresholds[class_name][idx], "max AP", list_ap[class_name][idx])

    for max_fp in [0.01, 0.03, 0.05, 0.1]:
      result[max_fp] = {"default": 0.5}
      for class_name in self.class_names:
        class_threshold = self.threshold(class_name, list_thresholds[class_name], list_ap[class_name], list_FP[class_name], max_fp)
        result[max_fp][class_name] = class_threshold

    return result

  def evaluate(self, confidence_thresholds={"default": 0.0}):
    """
    Non-classical AP-related FP-constrained AP@IoU=0.5 calculation
    """

    aps = []
    class_ap = {}
    class_FP = {}

    info("Evaluating non-classical FP-constrained AP@IoU=0.5")

    for class_name in self.class_names:
      ap, FP = self.compute_ap(class_name, confidence_thresholds.get(class_name, confidence_thresholds["default"]))
      aps.append(ap)
      class_ap[class_name] = ap
      class_FP[class_name] = FP

    mAP = np.mean(aps)
    return {'mAP': mAP, 'AP per class': class_ap, "FP per class": class_FP}



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import time

from sciveo.tools.logger import *
from sciveo.tools.configuration import ConfigurationArguments
from sciveo.monitoring.monitor import BaseMonitor


class MonitorStart:
  def __init__(self, **kwargs):
    self.arguments = ConfigurationArguments({
      "period": 120,
      "block": True,
      "fork": False,
      "fork_type": 0,
      "output_path": None,
    }, **kwargs)

    if len(kwargs) == 0:
      info("sciveo monitoring default options", self.arguments)

  def __call__(self):
    if self.arguments["fork"] and self.arguments["block"]:
      self.fork()
    self.start()

  def start_multiprocessing(self):
    self.start()
  def start_fork(self):
    os.setsid()
    self.start()
  def fork(self):
    if self.arguments["fork_type"] == 0:
      import multiprocessing
      background_process = multiprocessing.Process(target=self.start_multiprocessing)
      background_process.daemon = True
      background_process.start()
      info("monitoring service started with", self.arguments)
      exit()
    else:
      pid = os.fork()
      if pid == 0:
        self.start_fork()
      else:
        info("sciveo monitor service started with pid", pid, self.arguments)
        exit()

  def start(self):
    period = max(self.arguments["period"], 5)
    mon = BaseMonitor(period=period, output_path=self.arguments["output_path"])
    mon.start()
    mon.join()

    if self.arguments["block"]:
      while(True):
        time.sleep(60)



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2025
#

import time
import unittest

from sciveo.tools.totp import *


class TestTOTP(unittest.TestCase):
  @classmethod
  def setUpClass(cls):
    cls.secret = AuthTOTP.new_private_key(length=32)
    cls.auth = AuthTOTP(cls.secret, digits=8)

  def test_pass(self):
    token = self.auth.next_token()
    self.assertTrue(self.auth.verify_token(token=token))

  def test_expiry_time_fail(self):
    auth = AuthTOTP(self.secret, interval=3, digits=8)
    token = auth.next_token()
    self.assertTrue(auth.verify_token(token=token))
    time.sleep(2)
    self.assertTrue(auth.verify_token(token=token))
    time.sleep(2)
    self.assertTrue(auth.verify_token(token=token))
    time.sleep(2)
    self.assertFalse(auth.verify_token(token=token))


if __name__ == '__main__':
  unittest.main()


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import cv2
import numpy as np

from sciveo.tools.common import *
from sciveo.media.pipelines.processors.base import *


class MotionVideoWriter(VideoWriterFFMPEG):
  def write(self, frame):
    self.out.writeFrame(frame[:,::-1])


class VideoMotionDetector(BaseProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)

    self.default.update({"CRF": 40, "frame_rate": 1})

    # TODO: Place md-related options into config
    self.fgbg = cv2.createBackgroundSubtractorMOG2(history=50, varThreshold=20, detectShadows=False)

  def init_video(self, frame):
    tag = "MOTION"
    video_local_path = self.add_suffix_to_filename(self.local_path, tag)
    self.next_content(self.media, tag, video_local_path, w=frame.shape[1], h=frame.shape[0])
    self.md_video = MotionVideoWriter(video_local_path, crf=str(self["CRF"]))

  def process(self, media):
    self.media = media
    self.local_path = media["local_path"]

    frame_rate = self["frame_rate"]

    cap = cv2.VideoCapture(self.local_path)
    ret, frame = cap.read()
    self.init_video(frame)

    n_frame = 0
    while(cap.isOpened()):
      n_frame += 1
      ret, frame = cap.read()
      if not ret:
        break

      try:
        if n_frame % frame_rate == 0:
          md_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)
          md_frame = cv2.blur(src=md_frame, ksize=(10,10))
          md_frame = self.fgbg.apply(md_frame)
          self.md_video.write(md_frame)
      except Exception as e:
        exception(e)

    cap.release()
    self.md_video.close()
    return self.media

  def content_type(self):
    return "video"

  def name(self):
    return "video-motion-detector"



#
# Stanislav Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact s.georgiev@softel.bg.
#
# 2025
#

import struct
import math
from sciveo.tools.logger import *


def to_u16(regs):
  return regs[0] & 0xFFFF

def to_s16(regs):
  v = regs[0] & 0xFFFF
  return v - 0x10000 if v & 0x8000 else v

def to_u32_le_words(regs):
  # device transmits U32 double-word as little-endian word order:
  # Example in doc: 0x01020304 transmitted as 03,04,01,02 => regs[0]=0x0304, regs[1]=0x0102
  return ((regs[1] & 0xFFFF) << 16) | (regs[0] & 0xFFFF)

def to_s32_le_words(regs):
  v = to_u32_le_words(regs)
  return v - 0x100000000 if v & 0x80000000 else v

def to_u64_be(regs):
  # U64 big-endian: regs[0] = highest word
  return ((regs[0] & 0xFFFF) << 48) | ((regs[1] & 0xFFFF) << 32) | ((regs[2] & 0xFFFF) << 16) | (regs[3] & 0xFFFF)

def to_s64_be(regs):
  v = to_u64_be(regs)
  return v - 0x10000000000000000 if v & (1 << 63) else v

def float_from_words_be(regs):
  # regs[0] = high word, regs[1] = low word -> bytes: [hi_hi, hi_lo, lo_hi, lo_lo]
  b = (int(regs[0]) & 0xFFFF).to_bytes(2, "big") + (int(regs[1]) & 0xFFFF).to_bytes(2, "big")
  return struct.unpack(">f", b)[0]

def float_from_words_le(regs):
  # little-endian word order: regs[0] = low word, regs[1] = high word
  # reconstruct big-endian 4 bytes by swapping words
  b = (int(regs[1]) & 0xFFFF).to_bytes(2, "big") + (int(regs[0]) & 0xFFFF).to_bytes(2, "big")
  return struct.unpack(">f", b)[0]

def to_float_best(regs):
  """
  Try decoding FLOAT taking into account vendor word-order quirks.
  Prefer little-endian word order (as documented for U32), but fall back
  to big-endian word order if the result looks more reasonable.
  """
  try:
    val_le = float_from_words_le(regs)
  except Exception:
    val_le = None
  try:
    val_be = float_from_words_be(regs)
  except Exception:
    val_be = None

  def good(v):
    return v is not None and math.isfinite(v)

  if good(val_le) and not good(val_be):
    return val_le
  if good(val_be) and not good(val_le):
    return val_be
  if good(val_le) and good(val_be):
    if abs(val_le) < abs(val_be):
      return val_le
    return val_be
  return None

# Known register map (protocol addresses as in documentation). Format: addr: (type, factor, name)
EMS300_REG_MAP = {
  8018: ("U32", 1, "LC Total Number"),
  8020: ("U32", 1, "Total Number of Available LC"),
  8022: ("FLOAT", 1000, "Total Rated Power of Energy Storage"),
  8026: ("U32", 0.1, "Actual SOC"),
  8038: ("S32", 1, "Energy Storage Power"),
  8040: ("U32", 1, "Available Time for Maximum Power Discharging"),
  8042: ("U32", 1, "Available Time for Maximum Power Charging"),
  8044: ("S32", 1, "Plant Active Power"),
  8046: ("U32", 0.001, "Grid Frequency"),
  8048: ("S32", 1, "Plant Reactive Power"),
  8050: ("U32", 1, "BusVoltage"),
  8052: ("S32", 0.001, "Power Factor"),
  8056: ("S32", 1, "Photovoltaic Power Rating"),
  8058: ("S32", 1, "PV Available Reactive Power"),
  8060: ("S32", 1, "PV Active Power"),
  8062: ("S32", 1, "PV Reactive Power"),
  8064: ("U32", 1, "PV Subarray Number"),
  8070: ("U32", 1, "SOC Upper Limit Level 1 Protection"),
  8072: ("U32", 1, "SOC Lower Limit Level 1 Protection"),
  8074: ("U32", 1, "Total Field Charge Capacity"),
  8076: ("U32", 1, "Total Field Discharge Capacity"),
  8092: ("U32", 1, "Available Time of Real-time Charging"),
  8094: ("U32", 1, "Available Time of Real-time Discharging"),
  8098: ("U16", 1, "Charging Complete"),
  8099: ("U16", 1, "Discharge Completed"),
  8100: ("U16", 1, "Charge Locked"),
  8101: ("U16", 1, "Discharge Locked"),
  8297: ("U16", 1, "Plant Stop in Running Status"),
  8298: ("U16", 1, "Plant Standby in Running Status"),
  8299: ("U16", 1, "Plant Discharge in Running Status"),
  8300: ("U16", 1, "Plant Charge in Running Status"),
  8301: ("U16", 1, "Plant Fault in Running Status"),
  8485: ("FLOAT", 1, "Residual Capacity of Energy Storage"),
  8487: ("U16", 0.1, "Battery Voltage"),
  8488: ("FLOAT", 1000, "Active Load"),
  8490: ("FLOAT", 1000, "Rated Capacity of Energy Storage"),
  8492: ("U16", 1, "System Operating Status"),
  8493: ("U64", 1, "Start and Stop Process Status Word"),
  8638: ("FLOAT", 1, "Total DC Side Power of ESS"),
  8726: ("U16", 1, "Number of On-grid Devices"),
  8727: ("U16", 1, "Number of Off-grid Devices"),
  8728: ("FLOAT", 1, "Energy Storage SOC"),
  8733: ("FLOAT", 1, "Total Active Power of Chargers in the Plant"),
  8735: ("FLOAT", 1, "Total Energy Delivered by Chargers in the Plant"),
  8752: ("FLOAT", 1, "Upper Limit of Available Active Power of Plant"),
  8754: ("FLOAT", 1, "Lower Limit of Available Active Power of Plant"),
  8756: ("FLOAT", 1, "Available Plant Reactive Power"),
  8758: ("FLOAT", 0.1, "Active Power of Energy Storage"),
  8760: ("FLOAT", 0.1, "Reactive Power of Energy Storage"),
  8762: ("U32", 0.1, "Maximum SOC"),
  8766: ("U32", 0.1, "Minimum SOC"),
  8768: ("U32", 0.1, "Maximum Charging Power Allowed"),
  8770: ("U32", 0.1, "Maximum Discharging Power"),
  8772: ("U32", 0.1, "ESS Daily Charge"),
  8774: ("U32", 0.1, "ESS Daily Discharge"),
  8776: ("U32", 0.1, "System Dischargeable Energy"),
  8778: ("U32", 0.1, "System Chargeable Capacity"),
  9082: ("FLOAT", 1, "Dynamic Allowable Charging Power"),
  9084: ("FLOAT", 1, "Dynamic Allowable Discharging Power"),
  9086: ("U16", 1, "ESS Operating Status"),
  9087: ("FLOAT", 1, "ESS Power Factor"),
  9089: ("FLOAT", 1, "ESS SOH"),
  9115: ("FLOAT", 1, "ESS Grid-Connection Point Meter Total Daily Charge"),
  9117: ("FLOAT", 1, "ESS Grid-Connection Point Meter Total Daily Discharge"),
  9119: ("U16", 1, "ESS Fault"),
  9120: ("U16", 1, "Energy Storage Alarm"),
  9121: ("U16", 1, "ESS Communication Failure"),
  9122: ("U16", 1, "ESS HVAC Fault"),
  9424: ("U16", 1, "Active Power Limit Ratio"),
  8102: ("U16", 1, "Communication Status"),
  8302: ("U16", 1, "Device Alarm"),
  8303: ("U16", 1, "Device Fault"),
  8347: ("U16", 1, "Device Model"),
  8341: ("FLOAT", 1, "Daily Yield, kWh"),
  8379: ("U16", 1, "SN Code"),
  8389: ("U16", 1, "ETH1 mac"),
  8421: ("U16", 1, "ETH2 mac"),
  8453: ("U16", 1, "ETH3 mac"),
  8730: ("U16", 1, "Serial Port Communication Interrupted"),
  8731: ("U16", 1, "TCP Communication Interrupted"),
  8732: ("U16", 1, "Host EMS Communication Fault"),
  10601: ("U16", 1, "DI1 Value"),
  8554: ("U16", 1, "Host and Client Status"),
  8555: ("U16", 1, "Cascading"),
  8700: ("U16", 1, "Debugging Mode"),
  8711: ("U16", 1, "ESS Connection Point"),
  8712: ("U16", 1, "PV Connection Point"),
  8682: ("FLOAT", 1, "Replacement Percentage"),
  8713: ("U16", 1, "Energy Storage Data Source"),
  9826: ("U16", 1, "Select Power Distribution Strategy"),
  8570: ("U16", 1, "Mode selection"),
  8780: ("U16", 1, "Time-of-Use Power Period 1 Time Type"),
  8781: ("U16", 1, "Start Time of Time of Use Power Period 1"),
  8782: ("U16", 1, "End Time of Time of Use Power Period 1"),
  8783: ("FLOAT", 1, "Power Setting of Time of Use Power Period 1"),
  8785: ("U16", 1, "Time-of-Use Power Period 1 Charging/Discharging Status"),
  8894: ("U16", 1, "Time-of-Use Power Period 20 Time Type"),
  8895: ("U16", 1, "Start Time of Time of Use Power Period 20"),
  8896: ("U16", 1, "End Time of Time of Use Power Period 20"),
  8897: ("FLOAT", 1, "Power Setting of Time of Use Power Period 20"),
  8899: ("U16", 1, "Time-of-Use Power Period 20 Charging/Discharging Status"),
  9425: ("U16", 1, "Time-of-Use Power Minute/Second Mode"),
  9426: ("U16", 1, "Time-of-Use Power Second-Mode Period 1 Time Type"),
  9427: ("U32", 1, "Time-of-Use Power Second-Mode Period 1 Start Time"),
  9429: ("U32", 1, "Time-of-Use Power Second-Mode Period 1 End Time"),
  9431: ("FLOAT", 1, "Time-of-Use Power Second-Mode Period 1 Power"),
  9433: ("U16", 1, "Time-of-Use Power Second-Mode Period 1 Charging/Discharging Status"),
  9578: ("U16", 1, "Time-of-Use Power Second-Mode Period 20 Time Type"),
  9579: ("U32", 1, "Time-of-Use Power Second-Mode Period 20 Start Time"),
  9581: ("U32", 1, "Time-of-Use Power Second-Mode Period 20 End Time"),
  9583: ("FLOAT", 1, "Time-of-Use Power Second-Mode Period 20 Power"),
  9585: ("U16", 1, "Time-of-Use Power Second-Mode Period 20 Charging/Discharging Status"),
  9123: ("FLOAT", 1, "Remotely Set Time-of-Use Power Period 1 Start Time"),
  9125: ("FLOAT", 1, "Remotely Set Time-of-Use Power Period 1 End Time"),
  9127: ("FLOAT", 1, "Remotely Set Time-of-Use Power Period 1 Power"),
  9417: ("FLOAT", 1, "Remotely Set Time-of-Use Power Period 50 Start Time"),
  9419: ("FLOAT", 1, "Remotely Set Time-of-Use Power Period 50 End Time"),
  9421: ("FLOAT", 1, "Remotely Set Time-of-Use Power Period 50 Power"),
  8103: ("U16", 1, "Active Power Control Running Status"),
  8547: ("FLOAT", 1000, "Setting Feedback Value of Active Power"),
  8560: ("FLOAT", 1, "Active Power Control Dead Zone"),
  8710: ("U16", 1, "Scheduled Object"),
  9423: ("U16", 1, "Control Priority"),
  8714: ("U16", 1, "Operation Mode upon Communication Interruption"),
  8565: ("U32", 1, "Communication Interruption Time"),
  8567: ("U32", 1, "Communication Interruption Counting Period"),
  8569: ("U16", 1, "Communication Interruption Counts"),
  8105: ("U32", 1, "AGC_PLAN_001"),
  8107: ("U32", 1, "AGC_PLAN_002"),
  8293: ("U32", 1, "AGC_PLAN_095"),
  8295: ("U32", 1, "AGC_PLAN_096"),
  8304: ("U16", 1, "AGC Allow Control"),
  8305: ("U16", 1, "AGC Remote/Local Status"),
  8306: ("U16", 1, "AVC Turn Up Lock"),
  8307: ("U16", 1, "AVC Turn Down Lock"),
  8308: ("U16", 1, "AVC Allow Control"),
  8309: ("U16", 1, "AVC Remote/Local Status"),
  8104: ("U16", 1, "Reactive Power Control Running Status"),
  8549: ("FLOAT", 1000, "Setting Feedback Value of Reactive Power"),
  8562: ("FLOAT", 1, "Reactive Power Control Dead Zone"),
  8090: ("S32", 0.001, "Local Power Factor Set Value"),
  8689: ("U16", 1, "Execute Strategy Only When ESS is Discharging"),
  8693: ("FLOAT", 1, "Charge Step Factor"),
  8695: ("FLOAT", 1, "Discharge Step Factor"),
  8697: ("U16", 1, "Adjustment Period"),
  8698: ("FLOAT", 1, "System Discharge Threshold"),
  8706: ("FLOAT", 1, "PV Active Power Ratio K1"),
  8708: ("FLOAT", 1, "PV Active Power Ratio K2"),
  9827: ("FLOAT", 1, "Regular Load Power"),
  9829: ("FLOAT", 1, "SOC Threshold to Stop Supplying Power to Charger"),
  9831: ("FLOAT", 1, "Proportion of Stable Loads"),
  9836: ("U16", 1, "VSG Parameter Management"),
  9837: ("FLOAT", 1, "Output Voltage Setpoint"),
  9839: ("FLOAT", 1, "Output Voltage Setpoint (Hz)"),
  9841: ("FLOAT", 1, "Active Power-Frequency Droop Coefficient"),
  9843: ("FLOAT", 1, "Reactive Power-Voltage Droop Coefficient"),
  9845: ("FLOAT", 1, "Sent Active Power Setpoint"),
  9847: ("FLOAT", 1, "Sent Reactive Power Setpoint"),
  8640: ("U16", 1, "Frequency Droop Control Enable/Disable"),
  8642: ("U16", 1, "Frequency Response Type"),
  8310: ("U16", 1, "Frequency Control Running Flag"),
  8311: ("U32", 1, "Primary Frequency Regulation Control"),
  8313: ("U16", 1, "Primary Frequency Regulation Exit Signal"),
  8314: ("U16", 1, "Number of FM Algorithm Segments"),
  8315: ("U32", 0.001, "Rated Frequency of Grid-connection Point"),
  8317: ("U32", 0.01, "Frequency Regulation Coefficient"),
  8319: ("U32", 0.001, "Frequency Regulation Dead Zone"),
  8321: ("U32", 0.001, "Maximum Active Power Output"),
  8323: ("U32", 0.001, "Maximum Active Power Absorbed"),
  8325: ("U32", 0.001, "Lower Limit of Active Power Limit Coefficient"),
  8327: ("U32", 0.001, "Upper Limit of Active Power Limit Coefficient"),
  8329: ("U32", 0.01, "Frequency Regulation II Regulation Coefficient"),
  8331: ("U32", 0.001, "Frequency Regulation II Dead Zone"),
  8333: ("U32", 0.01, "Active Power Before Primary Frequency Regulation"),
  8335: ("U32", 0.01, "Active Power After Primary Frequency Regulation"),
  8645: ("FLOAT", 1, "Contingency FCAS Action Deadband fd"),
  8647: ("FLOAT", 1, "Contingency FCAS Maximum Output Active Power PA"),
  8649: ("FLOAT", 1, "Contingency FCAS Maximum Absorption Active Power PB"),
  8651: ("FLOAT", 1, "Contingency FCAS Frequency Deviation Threshold for Full-Scale Response"),
  8655: ("U16", 1, "Demand Control"),
  8656: ("U16", 1, "Select Strategy"),
  8657: ("U16", 1, "Allow ESS to Discharge or Not"),
  8658: ("FLOAT", 1, "Max. Allowable Demand"),
  8660: ("FLOAT", 1, "Lower Limit for Demand Control"),
  8662: ("U16", 1, "Reverse Power Protection"),
  8663: ("U16", 1, "Allow ESS charging during reverse power flow"),
  8664: ("FLOAT", 1, "Target Active Power of Reverse Power Protection"),
  8337: ("U32", 1, "Transformer Capacity"),
  8339: ("FLOAT", 1, "Active Power Conversion Coefficient"),
  8653: ("U16", 1, "Overload Protection"),
  8654: ("U16", 1, "Dynamic Capacity Expansion"),
  8666: ("U16", 1, "Power Backup"),
  8679: ("FLOAT", 1, "Recharging Power"),
  9833: ("FLOAT", 1, "K Value When Prioritizing ESS Charge"),
  8066: ("U32", 1, "SOC Upper Limit"),
  8068: ("U32", 1, "SOC Lower Limit"),
  8690: ("U32", 1, "SOC Balancing"),
  8667: ("FLOAT", 1, "SOC Upper Limit Level 1 Protection"),
  8669: ("FLOAT", 1, "SOC Upper Limit Level 1 Recovery"),
  8671: ("FLOAT", 1, "SOC Lower Limit Level 1 Recovery"),
  8673: ("FLOAT", 1, "SOC Lower Limit Level 1 Protection"),
  8675: ("FLOAT", 1, "SOC Lower Limit Level 2 Recovery"),
  8677: ("FLOAT", 1, "SOC Lower Limit Level 2 Protection"),
  8553: ("U16", 1, "Grid-connected/Off-grid Startup"),
  8715: ("U16", 1, "Inverter Operating Mode"),
  8718: ("FLOAT", 1, "Black Start SOC Threshold"),
  8720: ("U16", 1, "Minimum Number of Running Units"),
  8721: ("FLOAT", 1, "Load Rated Power"),
  8723: ("U16", 1, "System Waiting time"),
  8724: ("U16", 1, "System Timeout"),
   8725: ("U16", 1, "Charging Initialization Mode"),
  9855: ("U16", 1, "Auto-start on Recovery from System Fault"),
  9854: ("U16", 1, "Function Mode"),
  10001: ("U16", 1, "Total Number of Connected Devices"),
  10002: ("U16", 1, "Total Number of Fault Count"),
  10101: ("S32", 0.01, "Total Active Power"),
  10103: ("S32", 0.01, "Gateway Meter Active Power"),
  10105: ("S32", 0.01, "Load Power"),
  10107: ("U32", 0.1, "Gateway Meter Import Energy"),
  10109: ("U32", 0.1, "Gateway Meter Export Energy"),
  10501: ("U16", 1, "Device Type Code"),
  10502: ("U32", 1, "Protocol No."),
  10504: ("U32", 1, "Protocol Version No."),
  10506: ("U32", 1, "DI Status"),
  10701: ("U32", 0.1, "Max. Total Rated Active Power"),
  10703: ("U32", 0.1, "Total Battery Rated Capacity"),
  10705: ("U32", 0.1, "Battery Charge/Discharge Maximum Description"),
  10707: ("U16", 0.1, "Maximum Battery Charge Power"),
  10708: ("U16", 0.1, "Minimum Battery Charge Power"),
  10709: ("U16", 0.1, "Maximum Battery Discharge Power"),
  10710: ("U16", 0.1, "Minimum Battery Discharge Power"),
  10713: ("S32", 0.01, "Battery Power"),
  10715: ("U16", 0.1, "Battery level (SOC)"),
  10716: ("U32", 0.1, "Daily Battery Charging Energy"),
  10718: ("U32", 0.1, "Daily Battery Discharging Energy"),
  10720: ("U64", 0.1, "Total Battery Charging Energy"),
  10724: ("U64", 0.1, "Total battery discharge"),
  10901: ("U16", 1, "Maximum Total PV Rated Active Power"),
  10902: ("U16", 1, "Minimum Total PV Rated Active Power"),
  10903: ("U16", 1, "Maximum Total PV Rated Reactive Power"),
  10904: ("S16", 1, "Minimum Total PV Rated Reactive Power"),
  10905: ("U16", 1, "Set Total Inverter Active Power"),
  10906: ("S16", 1, "Set Total Inverter Reactive Power"),
  10909: ("S64", 1, "P-total-PV"),
  10913: ("U32", 0.1, "Daily PV yield"),
  10915: ("S64", 1, "Q-total-PV"),
  10919: ("U64", 0.1, "Total PV Yield"),
  10923: ("U32", 0.1, "PV Min. adjustable active power"),
  10925: ("U32", 0.1, "PV Max. adjustable active power"),
  10927: ("S32", 0.1, "PV Min. Adjustable Reactive Power"),
  10929: ("S32", 0.1, "PV Max. Adjustable Reactive Power"),
  10931: ("S32", 0.1, "PV Rated Active Power"),
  10933: ("S32", 0.1, "PV Rated Reactive Power"),
  10935: ("U16", 1, "Total Grid-Connected Inverter Devices"),
  10936: ("U16", 1, "Total Off-Grid Inverter Devices"),
  10937: ("U64", 0.1, "Monthly PV Yield"),
  10941: ("U64", 0.1, "PV Annual Yield"),
  10945: ("U64", 1, "PV Apparent Power"),
  10949: ("U16", 0.1, "PV Active Power Dispatching Ratio"),
  11804: ("U16", 1, "2030.5 Feed-in Control Method"),
  11805: ("S32", 0.01, "2030.5 Feed-in Limitation Value"),
  11807: ("S16", 0.1, "2030.5 Feed-in Limitation Ratio"),
}

def count_for_type(t):
  return {"U16": 1, "S16": 1, "U32": 2, "S32": 2, "U64": 4, "S64": 4, "FLOAT": 2}.get(t, 1)

def decode_registers(regs, dtype):
  if dtype == "U16":
    return to_u16(regs)
  if dtype == "S16":
    return to_s16(regs)
  if dtype == "U32":
    return to_u32_le_words(regs)
  if dtype == "S32":
    return to_s32_le_words(regs)
  if dtype == "U64":
    return to_u64_be(regs)
  if dtype == "S64":
    return to_s64_be(regs)
  if dtype == "FLOAT":
    return to_float_best(regs)
  return None

def read_input_registers(client, REG_MAP, addr, device_id):
  try:
    dtype, factor, name = REG_MAP[addr]
    count = count_for_type(dtype)
    protocol_addr = addr - 1
    rr = client.read_input_registers(address=protocol_addr, count=count, device_id=device_id)
    if rr is None:
      warning(f"No response for {addr} ({name})")
      return None
    if hasattr(rr, "isError") and rr.isError():
      warning(f"Modbus exception at {addr} ({name}): {rr}")
      return None
    regs = getattr(rr, "registers", None)
    if not regs or len(regs) < count:
      warning(f"Incomplete registers at {addr} ({name}): {regs}")
      return None
    val_raw = decode_registers(regs, dtype)
    if val_raw is None:
      warning(f"Unable to decode {addr} ({name}) regs={regs}")
      return None
    value = val_raw * factor
    # format integer-looking floats without trailing .0
    if isinstance(value, float) and value.is_integer():
      value = int(value)
    debug(f"{addr} {name}: raw={regs} -> {value} (factor={factor}, type={dtype})")
  except Exception as e:
    exception(e, f"read_input_registers::reading {addr} ({name})")
    return None
  return value, name, dtype



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2025
#

import time
import psutil

from sciveo.tools.logger import *
from sciveo.monitoring.watchdog.base import BaseWatchDogDaemon


class ProcessWatchDogDaemon(BaseWatchDogDaemon):
  def __init__(self, pid=None, process_cmd=None, period=5, command="echo '⚠️ process error!'", thread_inactive_threshold=None):
    self.process_cmd = process_cmd
    self.info_threads = {}
    self.thread_inactive_threshold = thread_inactive_threshold
    super().__init__(threshold_percent=10, period=period, command=command, monitored="Process", monitor_value="is dead")

    self.this_pid = psutil.Process().pid
    self.pids = self.get_pids_by_cmd()
    if pid is not None:
      self.pids.append(pid)

  def get_pids_by_cmd(self):
    matched_pids = []
    if self.process_cmd is not None:
      for proc in psutil.process_iter(attrs=["pid", "cmdline"]):
        if proc.info["pid"] == self.this_pid:
          continue
        try:
          cmdline = " ".join(proc.info["cmdline"])
          if self.process_cmd in cmdline:
            matched_pids.append(proc.info["pid"])
        except Exception:
          continue
      info(f"processes matching [{self.process_cmd}]", matched_pids)
    return matched_pids

  def get_threads(self, process, pid):
    current_time = time.time()
    for t in process.threads():
      cpu_time = t.user_time + t.system_time

      self.info_threads.setdefault(pid, {})
      self.info_threads[pid].setdefault(t.id, {
        "id": t.id,
        "start_at": current_time,
        "last_active": current_time,
        "last_cpu": cpu_time,
        "last_check": current_time
      })

      if cpu_time != self.info_threads[pid][t.id]["last_cpu"]:
        self.info_threads[pid][t.id]["last_active"] = current_time
      self.info_threads[pid][t.id]["last_cpu"] = cpu_time
      self.info_threads[pid][t.id]["last_check"] = current_time

  def check_threads(self, process, pid):
    current_time = time.time()
    delete_thread_info = []
    for tid, thr in self.info_threads[pid].items():
      check_time = int(current_time - thr["last_check"])
      if check_time > 1.5 * self.period:
        delete_thread_info.append(tid)
        continue
      elapsed_time = int(current_time - thr["last_active"])
      if elapsed_time > self.thread_inactive_threshold:
        warning(f"⚠️ process[{pid}] thread[{tid}] inactive for {elapsed_time}s.")
    for tid in delete_thread_info:
      self.info_threads[pid].pop(tid, None)

  def process_value(self, pid):
    try:
      p = psutil.Process(pid)
      if p.is_running():
        if p.status() != psutil.STATUS_ZOMBIE:

          if self.thread_inactive_threshold is not None:
            try:
              self.get_threads(p, pid)
              self.check_threads(p, pid)
            except psutil.NoSuchProcess as e:
              exception(e, pid)

          return 0
        else:
          return 50
    except psutil.NoSuchProcess:
      pass
    return 100

  def value(self):
    pv = 0.0
    for pid in self.pids:
      pv += self.process_value(pid)
    return {"percent": pv}


if __name__ == '__main__':
  daemons = [
    ProcessWatchDogDaemon(pid=87443, thread_inactive_threshold=10),
    ProcessWatchDogDaemon(process_cmd="sciveo monitor", thread_inactive_threshold=10),
  ]
  for d in daemons:
    d.start()

  while(True):
    time.sleep(30)


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import cv2
import gc
from PIL import Image

import torch
from transformers import AutoProcessor, AutoModelForCausalLM

from sciveo.tools.common import *
from sciveo.ml.images.description import ImageToText


class VideoToText(ImageToText):
  def __init__(self, model_id, max_length=64, cache_dir=None, device=None) -> None:
    super().__init__(model_id, max_length, cache_dir, device)

  def read_video_frame(self, video_path):
    cap = cv2.VideoCapture(video_path)
    total_frames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    cap.set(cv2.CAP_PROP_POS_FRAMES, total_frames // 2)
    ret, frame = cap.read()
    frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
    cap.release()
    return frame

  def predict(self, X):
    frames = []
    for x in X:
      frames.append(self.read_video_frame(x))
    predictions = super().predict(frames)
    return predictions



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import cv2
import gc
from PIL import Image

import torch
from diffusers import StableDiffusionPipeline, LMSDiscreteScheduler
from diffusers import StableDiffusionInstructPix2PixPipeline, EulerAncestralDiscreteScheduler
from transformers import AutoProcessor, AutoModelForCausalLM
from transformers import Blip2Processor, Blip2ForConditionalGeneration

from sciveo.tools.logger import *
from sciveo.tools.common import *
from sciveo.media.pipelines.processors.tpu_base import *
from sciveo.media.pipelines.base import ApiContent
from sciveo.ml.images.description import ImageToText


class ImageDiffusionText(TPUBaseProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)
    self.default.update({
      "prompt": "monkeys eating bananas and watching football",
      "iterations": 50,
      "model_version": 1,
      "dtype": 1,
      "width": 512,
      "height": 512,
      "num_images": 1,
      "bucket": "smiveo-images"
    })

    self.max_batch_size = 3 # GPU memory dependent
    self.width = self.get("width", 128, 512)
    self.height = self.get("height", 128, 512)

    cache_dir = os.path.join(os.environ['MEDIA_MODELS_BASE_PATH'], "models/")

    model_names = ["softel/stable-diffusion-v1.0", "softel/stable-diffusion-v1.1"]
    torch_dtypes = [torch.float16, torch.float32]

    model_name = model_names[max(0, min(len(model_names) - 1, self["model_version"]))]
    torch_dtype = torch_dtypes[max(0, min(len(torch_dtypes) - 1, self["dtype"]))]
    TPU = os.environ.get("MEDIA_PROCESSING_BACKEND", "cpu")

    debug("init", model_name, torch_dtype, "on", TPU)

    # TODO: Add scheduler config param
    lms = LMSDiscreteScheduler(
      beta_start=0.00085,
      beta_end=0.012,
      beta_schedule="scaled_linear"
    )

    self.pipe = StableDiffusionPipeline.from_pretrained(
      model_name,
      scheduler=lms,
      torch_dtype=torch_dtype,
      cache_dir=cache_dir
    ).to(TPU)

    self.progress_per_media = self.max_batch_size * self.max_progress / self["iterations"] / self["num_images"]

  def on_progress(self, step, timestep, latents):
    debug("progress", step)
    MediaJobState.queue().inc_progress(self.job_id, self.progress_per_media)

  def generate(self, image_local_paths):
    image_samples = self.pipe(
      prompt=self["prompt"],
      width=self.width,
      height=self.height,
      num_inference_steps=self.get("iterations", 10, 300),
      num_images_per_prompt=len(image_local_paths),
      callback_steps=1,
      callback=self.on_progress
    )

    for i, image in enumerate(image_samples["images"]):
      debug("generate save", i, image_local_paths[i])
      image.save(image_local_paths[i])

    del image_samples

  def run(self, job, input):
    self.job_id = job["id"]
    parent_guid = job['content_id']

    next_media = {
      "media": [],
      "guid": [],
      "key": [],
      "local_path": []
    }

    for i in range(self.get("num_images", 1, 33)):
      # guid = f"{self.name()}-{datetime.datetime.now().strftime('%Y%m%d%H%M%S')}-{parent_guid}"
      guid = self.new_guid()
      key = f"{guid}.jpg"
      next_media["key"].append(key)
      next_media["guid"].append(guid)
      next_media["local_path"].append(os.path.join(self.base_tmp_path, key))

    list_local_paths = [next_media["local_path"][i : i + self.max_batch_size] for i in range(0, len(next_media["local_path"]), self.max_batch_size)]
    debug("run list_local_paths", list_local_paths)
    for local_paths in list_local_paths:
      self.generate(local_paths)

    for i, guid in enumerate(next_media["guid"]):
      media = {
        "guid": guid,
        "parent": parent_guid,
        "content_type": "image",
        "owner": job["owner"],
        "name": f"[{self['prompt']}] {self['iterations']}",
        "local_path": next_media["local_path"][i],
        "w": self.width, "h": self.height,
        "height": self.height,
        "key": next_media["key"][i],
        "bucket": self["bucket"],
        "processor": self.name(),
        "layout": {"name": self.name(), "height": self.height, **self["layout"]}
      }

      if self["output"]:
        job["output"].append(media)
      if self["append-content"]:
        job["append-content"].append(media)
      next_media["media"].append(media)

    return next_media["media"]

  def content_type(self):
    return "image"

  def name(self):
    return "image-diffusion"

  def is_append_processor(self):
    return False


class ImageToTextProcessor(TPUBaseProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)

    self.api = ApiContent()

    self.cache_dir = os.path.join(os.environ['MEDIA_MODELS_BASE_PATH'], "models/")
    self.device = os.environ.get("MEDIA_PROCESSING_BACKEND", "cpu")

    self.default.update({
      "max_length": 64,
      "model_id": 0,
      "output": False
    })

    self.predictor = None

  def process(self, media):
    debug("process", media['guid'])
    if self.predictor is None:
      self.predictor = ImageToText(self['model_id'], self['max_length'], self.cache_dir, self.device)
    local_path = media["local_path"]
    predict = self.predictor.predict_one(local_path)
    return self.set_media(media, predict)

  def set_media(self, media, predict):
    media.setdefault("next", [])
    media["next"].append({
      "guid": f"TXT-{media['guid']}",
      "parent": media['guid'],
      "content_type": "comment",
      "content_text": predict,
      "owner": media["owner"],
      "name": f"{predict} [{self['max_length']}]",
      "processor": self.name()
    })

    self.api.update(media, {"description": predict})

    return media

  def content_type(self):
    return "image"

  def name(self):
    return "image-to-text"

  def is_append_processor(self):
    return True


class ImageDiffusionImageText(TPUBaseProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)
    self.default.update({
      "prompt": "monkeys eating bananas and watching football",
      "iterations": 10,
      "dtype": 0,
      "max_dim": 512,
      # "num_images": 1,
      "bucket": "smiveo-images"
    })

    self.max_batch_size = 1 # GPU memory dependent
    self.max_dim = self.get("max_dim", 128, 512)

    cache_dir = os.path.join(os.environ['MEDIA_MODELS_BASE_PATH'], "models/")

    torch_dtypes = [torch.float16, torch.float32]
    model_name = "softel/image-instruct-v1.0"
    torch_dtype = torch_dtypes[max(0, min(len(torch_dtypes) - 1, self["dtype"]))]
    TPU = os.environ.get("MEDIA_PROCESSING_BACKEND", "cpu")

    debug("init", model_name, torch_dtype, "on", TPU)

    self.pipe = StableDiffusionInstructPix2PixPipeline.from_pretrained(
      model_name,
      torch_dtype=torch_dtype,
      safety_checker=None,
      cache_dir=cache_dir
    ).to(TPU)

    self.pipe.scheduler = EulerAncestralDiscreteScheduler.from_config(self.pipe.scheduler.config)

    self.progress_per_media = self.max_progress / self["iterations"]
    # self.progress_per_media = self.max_batch_size * self.max_progress / self["iterations"] / self["num_images"]

  def on_progress(self, step, timestep, latents):
    debug("progress", step)
    MediaJobState.queue().inc_progress(self.job_id, self.progress_per_media)

  def process(self, media):
    try:
      self.media = media
      self.local_path = self.media["local_path"]
      frame = cv2.imread(self.local_path)

      tag = f"IT2I-{self['prompt'].replace(' ','')[:16]}"
      image_local_path = self.add_suffix_to_filename(self.local_path, tag)

      h, w = get_frame_resolution_max_dim(frame, self.max_dim)
      frame_resized = cv2.resize(frame, (w, h), interpolation=cv2.INTER_AREA)

      image_samples = self.pipe(
        self["prompt"],
        image=Image.fromarray(cv2.cvtColor(frame_resized, cv2.COLOR_BGR2RGB)),
        image_guidance_scale=1,
        num_inference_steps=self.get("iterations", 3, 100),
        num_images_per_prompt=1,
        callback=self.on_progress
      )

      image_samples["images"][0].save(image_local_path)
      del image_samples

      self.next_content(self.media, tag, image_local_path, w=frame_resized.shape[1], h=frame_resized.shape[0], name=f"[{self['prompt']}] {self['iterations']}")
    except Exception as e:
      exception(e, self.media)
    return self.media

  def content_type(self):
    return "image"

  def name(self):
    return "image-diffusion-image-text"

  def is_append_processor(self):
    return False



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os

from sciveo.tools.logger import *
from sciveo.tools.synchronized import DataSynchronized
from sciveo.tools.http import *
from sciveo.tools.simple_counter import Timer


class PipelineQueue(DataSynchronized):
  def __init__(self, tag=""):
    super().__init__(tag)
    self.timer = Timer()

  def update_job_state(self, job_id, state):
    url = f"{os.environ['MEDIA_API_BASE_URL']}/api/v1/media_pipeline_job/state/?job_id={job_id}&transit_to={state}"
    headers = { "Auth-Token": os.environ['MEDIA_API_AUTH_TOKEN'] }
    response = GET(url, headers=headers)

  def update_job_progress(self, job_id, progress):
    elapsed = self.timer.stop()
    debug("progress", progress, self.data_sync[job_id]["progress"], "elapsed", elapsed)
    if elapsed > 1:
      progress = int(min(progress, 100.0))
      url = f"{os.environ['MEDIA_API_BASE_URL']}/api/v1/media_pipeline_job/progress/?job_id={job_id}&set={progress}"
      headers = { "Auth-Token": os.environ['MEDIA_API_AUTH_TOKEN'] }
      response = GET(url, headers=headers)
      self.timer.start()

  def start(self, job_id):
    with self.lock_data:
      self.data_sync[job_id] = {"state": "started", "progress": 0}
    self.update_job_progress(job_id, 0)
    self.update_job_state(job_id, "started")

  def inc_progress(self, job_id, inc_value):
    progress = 0
    with self.lock_data:
      self.data_sync[job_id]["progress"] += inc_value
      self.data_sync[job_id]["progress"] = min(self.data_sync[job_id]["progress"], 100)
      progress = self.data_sync[job_id]["progress"]
    self.update_job_progress(job_id, progress)

  def finish(self, job_id, job):
    with self.lock_data:
      self.data_sync[job_id] = {"state": "finished", "progress": 100, "job": job}
    self.update_job_progress(job_id, 100)
    self.update_job_state(job_id, "finished")


class MediaJobState:
  _queue = None

  @staticmethod
  def queue():
    if MediaJobState._queue is None:
      MediaJobState._queue = PipelineQueue("media jobs")
    return MediaJobState._queue



import math
import cv2
import numpy as np

from geometrics import *
from laser import *
from volume_bar import *
from base_move import *


class RocketBase(BaseMoveConstantVelocity):
  def __init__(self, x0, y0):
    super().__init__(x0, y0, 0, 0, (255, 0, 0))
    self.l = 10000000000
    self.angle = 0

  def check_collision(self, x, y, r):
    return False

  def rotate(self, angle):
    self.angle += angle

  def accelerate(self, a):
    pass

  def fire(self):
    pass

  def alive(self):
    return True

  def laser_positions(self):
    return []


class SimpleRocket(RocketBase):
  def __init__(self, x0=0, y0=0, color=(255, 0, 0)):
    super().__init__(x0, y0)
    self.color = color
    self.r = 10

    self.vx = 0
    self.vy = 0

  def accelerate(self, gx, gy):
    self.vx += gx
    self.vy += gy

  def draw(self, frame):
    cv2.circle(frame, (int(self.x), int(self.y)), self.r, self.color, thickness=-1)
    cv2.putText(frame, "V ({},{})".format(round(self.vx, 2), round(self.vy, 2)), (5, 20), fontFace=cv2.FONT_HERSHEY_SIMPLEX, fontScale=0.5, color=(255, 255, 255), thickness=1)


class Rocket(RocketBase):
  def __init__(self, x0=0, y0=0, color=(255, 0, 0), txt_pos=(5, 20), l=10, laser_count=100, fuel=20):
    super().__init__(x0, y0)
    self.color = color
    self.txt_pos = txt_pos
    self.l = l

    self.thruster_default = 30
    self.thruster = 0
    self.fuel = fuel

    self.a = 0
    self.ax = 0
    self.ay = 0

    self.vx = 0
    self.vy = 0

    self.laser_count = laser_count
    self.lasers = []

    self.fuel_bar = VolumeBar((400, 20), self.fuel, name="FUEL", color=(255, 0, 0))
    self.laser_bar = VolumeBar((400, 50), self.laser_count, name="LASER", color=(0, 0, 255))

  def accelerate(self, a):
    if self.fuel < a:
      a = self.fuel
    self.fuel -= a

    angle = to_radians(self.angle)
    self.ax = a * math.cos(angle)
    self.ay = a * math.sin(angle)
    self.a = a

    self.vx += self.ax
    self.vy += self.ay

    self.thruster = self.thruster_default

  def draw(self, frame):
    angle = to_radians(self.angle)
    p0 = (int(self.x), int(self.y))
    p1 = (int(self.x + self.l * math.cos(angle)), int(self.y + self.l * math.sin(angle)))
    p2 = (int(self.x - self.l * math.cos(angle)), int(self.y - self.l * math.sin(angle)))
    p3, p4 = line_perpendicular(p1, p2, self.l)

    cv2.line(frame, p1, p2, self.color, 2)
    # cv2.line(frame, p3, p4, self.color, 4)
    cv2.line(frame, p3, p1, (0, 0, 255), 3)
    cv2.line(frame, p4, p1, (0, 255, 0), 3)
    cv2.line(frame, p3, p0, self.color, 2)
    cv2.line(frame, p4, p0, self.color, 2)

    if self.thruster > 0:
      l = self.a * 200
      p1 = p2
      da = 0.3
      p2 = [int(p1[0] - l * math.cos(angle - da)), int(p1[1] - l * math.sin(angle - da))]
      p3 = [int(p1[0] - l * math.cos(angle + da)), int(p1[1] - l * math.sin(angle + da))]
      cv2.drawContours(frame, [np.array([p1, p2, p3])], 0, (0, 0, 255), -1)
      self.thruster -= 1

    txt = "V ({},{}) A[{}]".format(round(self.vx, 3), round(self.vy, 3), self.angle)
    cv2.putText(frame, txt, self.txt_pos, fontFace=cv2.FONT_HERSHEY_SIMPLEX, fontScale=0.3, color=(255, 255, 255), thickness=1)

    self.move_lasers(frame)

    self.fuel_bar(frame, self.fuel)
    self.laser_bar(frame, self.laser_count)

  def move_lasers(self, frame):
    for laser in self.lasers:
      if not laser.move(frame):
        self.lasers.remove(laser)

  def fire(self):
    if self.laser_count > 0:
      self.laser_count -= 1
      self.lasers.append(Laser(self.x, self.y, self.angle))

  def laser_positions(self):
    positions = []
    for laser in self.lasers:
      positions.append([laser.p1, laser.p2, laser.angle])
    return positions

  def remove_laser(self, idx):
    if idx < len(self.lasers):
      del self.lasers[idx]

  def check_collision(self, x, y, r):
    # d = np.linalg.norm(np.array([x, y]) - np.array([self.x, self.y]))
    d = math.sqrt((x - self.x) ** 2 + (y - self.y) ** 2)
    return d < r + self.l


class RocketExplosion(RocketBase):
  def __init__(self, x0, y0):
    super().__init__(x0, y0)
    self.r = 3

  def move(self, frame):
    cv2.circle(frame, (int(self.x), int(self.y)), int(self.r), (0, 0, 255), thickness=-1)
    self.r += 0.8

  def alive(self):
    return self.r < 100


import cv2
import math
import numpy as np

from geometrics import to_radians
from base_move import *


class Laser(BaseMoveConstantVelocity):
  def __init__(self, x0, y0, angle):
    super().__init__(x0, y0, 0, 0, (0, 0, 255))
    # self.path = 0
    # self.x0 = x0
    # self.y0 = y0

    self.l = 10
    self.angle = angle

    self.points()

    self.v = 2.5
    self.vx = self.v * math.cos(to_radians(self.angle))
    self.vy = self.v * math.sin(to_radians(self.angle))

  def points(self):
    angle = to_radians(self.angle)
    self.p1 = (int(self.x), int(self.y))
    self.p2 = (int(self.x + self.l * math.cos(angle)), int(self.y + self.l * math.sin(angle)))
    # self.path = math.sqrt((self.x - self.x0) ** 2 + (self.y - self.y0) ** 2)

  def move(self, frame):
    w = frame.shape[1]
    h = frame.shape[0]

    self.x += self.vx
    self.y += self.vy

    self.points()

    cv2.line(frame, self.p1, self.p2, self.color, 2)

    return self.x >= 0 and self.x <= w and self.y >= 0 and self.y <= h


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#


class BaseEncoder:
  def fit(self, X):
    pass

  def transform(self, X):
    pass

  def fit_transform(self, X):
    pass

  def inverse(self, X):
    pass


import math
import cv2


class VolumeBar:
  def __init__(self, p1, max_vol, name, color):
    self.w = 100
    self.h = 10
    self.p1 = p1
    self.max_vol = max_vol
    self.name = name
    self.color = color

    self.p2 = [p1[0] + self.w, p1[1] + self.h]

  def __call__(self, frame, vol):
    self.draw(frame, vol)

  def draw(self, frame, vol):
    vol = min(self.max_vol, vol)
    p3 = [int(self.p1[0] + self.w * vol / self.max_vol), self.p1[1] + self.h]
    cv2.putText(frame, self.name, (self.p2[0] + 3, self.p2[1]), fontFace=cv2.FONT_HERSHEY_SIMPLEX, fontScale=0.5, color=(255, 255, 255), thickness=1)
    cv2.rectangle(frame, self.p1, p3, self.color, -1)
    cv2.rectangle(frame, self.p1, self.p2, (255, 255, 255), 1)



import cv2

class GravityBall:
  def __init__(self, r, x=0, y=0, gx=0.2, gy=0.1, color=(255, 0, 0)):
    self.color = color
    self.r = r

    self.x = 0
    self.y = 0

    self.vx = 0
    self.vy = 0
    self.gx = gx
    self.gy = gy

  def move(self, frame):
    w = frame.shape[1]
    h = frame.shape[0]

    self.vx += self.gx
    self.vy += self.gy
    self.x += self.vx
    self.y += self.vy

    if self.x + self.r >= w:
      self.x = w - self.r
      self.vx = - self.vx
    if self.y + self.r >= h:
      self.y = h - self.r
      self.vy = - self.vy

    cv2.circle(frame, (int(self.x), int(self.y)), self.r, self.color, thickness=-1)


#!/usr/bin/env python
#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import time
import json
import argparse

from sciveo.version import __version__
from sciveo.tools.logger import *
from sciveo.tools.configuration import GlobalConfiguration


def main():
  config = GlobalConfiguration.get()

  parser = argparse.ArgumentParser(description=f'{config.name} CLI')
  parser.add_argument(
    'command',
    choices=[
      'init', 'monitor', 'scan',
      'nvr', 'rtsp', 'capture',
      'predictors-server', 'media-server', 'media-run',
      'watchdog',
    ],
    help='Command to execute')

  parser.add_argument('--period', type=int, default=120, help='Period in seconds')
  parser.add_argument('--block', type=bool, default=True, help='Block flag')
  parser.add_argument('--auth', type=str, default=config['secret_access_key'], help='Auth secret access key')
  parser.add_argument('--timeout', type=float, default=1.0, help='Timeout')
  parser.add_argument('--net', type=str, default=None, help='Network like 192.168.10.0/24')
  parser.add_argument('--url', type=str, default=None, help='URL')
  parser.add_argument('--host', type=str, default=None, help='Host ip or name')
  parser.add_argument('--port', type=int, default=22, help='Host port number, used for network ops')
  parser.add_argument('--ports', type=str, default="[]", help='Host ports list')
  parser.add_argument('--localhost', type=bool, default=False, help='Add localhost to list of hosts')
  parser.add_argument('--input-path', type=str, default=None, help='Input Path')
  parser.add_argument('--output-path', type=str, default=None, help='Output Path')
  parser.add_argument('--width', type=int, default=None, help='width')
  parser.add_argument('--height', type=int, default=None, help='height')
  parser.add_argument('--fps', type=int, help='FPS')
  parser.add_argument('--stream', type=str, default=None, help='Stream')
  parser.add_argument('--rate', type=int, help='Rate number')
  parser.add_argument('--processor', type=str, help='Processor name')
  parser.add_argument('--src', type=str, default=None, help='Source')
  parser.add_argument('--src-id', type=int, default=None, help='Source Id')
  parser.add_argument('--dst', type=str, default=None, help='Destination')
  parser.add_argument('--value', type=float, help='Value')
  parser.add_argument('--threshold', type=float, default=None, help='Threshold')
  parser.add_argument('--execute', type=str, default=None, help='Execute command')
  parser.add_argument('--pid', type=int, default=None, help='Process PID')
  parser.add_argument('--serial', type=str, default=None, help='Serial Name')
  parser.add_argument('--data-json', type=str, default="{}", help='Json Data')
  parser.add_argument('--display', type=int, default=None, help='Display number')

  args = parser.parse_args()

  if args.command == 'monitor':
    if args.src is None:
      from sciveo.monitoring.start import MonitorStart
      MonitorStart(period=args.period, block=args.block, output_path=args.output_path)()
    elif args.src.startswith("power"):
      from sciveo.monitoring.power.ems300 import PowerEMS300
      mon = PowerEMS300(serial=args.serial, host=args.host, port=args.port, period=args.period, output_path=args.output_path)
      mon.start()
      while(True):
        time.sleep(3600)
  elif args.command == 'scan':
    from sciveo.network.tools import NetworkTools
    host=args.host
    if host is None:
      NetworkTools(timeout=args.timeout, localhost=args.localhost).scan_port(port=args.port, network=args.net)
    else:
      NetworkTools(timeout=args.timeout, ports=json.loads(args.ports)).scan_host(host)
  elif args.command == 'init':
    info(f"init {config.name} ver {__version__}")
    home = os.path.expanduser('~')
    base_path = os.path.join(home, f'.{config.name}')
    if not os.path.exists(base_path):
      os.makedirs(base_path)
      default_lines = [
        "secret_access_key=<your secret access key>",
        f"api_base_url=https://{config.name}.com",
        "sci_log_level=DEBUG"
      ]
      with open(os.path.join(base_path, "default"), 'w') as fp:
        for line in default_lines:
          fp.write(line + '\n')
    else:
      info(f"init, [{base_path}] already there")
  elif args.command == 'nvr':
    from sciveo.media.capture.nvr import VideoRecorder
    VideoRecorder(args.input_path).start()
  elif args.command == 'rtsp':
    if args.src is None and args.src_id is None:
      from sciveo.media.capture.nvr import RTSPVideoPlayer
      if args.url is not None:
        url = args.url
      elif args.host is not None and args.stream is not None:
        url = f"rtsp://{args.host}:{args.port}/{args.stream}"
      else:
        warning("Invalid URL")
        return
      player = RTSPVideoPlayer(url)
      player.run()
    else:
      from sciveo.media.capture.gst_server import CamFactory, ColorBarFactory, CamPlayerDaemon
      data = json.loads(args.data_json)
      if args.src_id is not None:
        server = CamFactory(args.src_id, args.width, args.height, args.fps, configuration=data)
      elif args.src is not None:
        if "color" in args.src:
          server = ColorBarFactory(args.width, args.height, args.fps)
        else:
          # warning("RTSP src", args.src, "not recognised")
          server = CamFactory(args.src, args.width, args.height, args.fps, configuration=data)
      if args.display is not None:
        player = CamPlayerDaemon(server, host=args.host, port=args.port, stream=args.stream, display=args.display)
        player.start()
        player.play()
      else:
        server.serve(host=args.host, port=args.port, stream=args.stream)
  elif args.command == 'capture':
    from sciveo.media.capture.cam import CameraDaemon, ScreenDaemon, CapturePlayer
    if args.src == "screen":
      cap = ScreenDaemon(src=args.src_id, region=json.loads(args.data_json))
    elif args.src == "cam":
      cap = CameraDaemon(cam_id=args.src_id)
    elif args.src == "rpi":
      from sciveo.media.capture.rpi import RPICamera
      cap = RPICamera(cam_id=args.src_id, width=args.width, height=args.height)
    else:
      cap = CameraDaemon(cam_id=args.src_id)
    player = CapturePlayer(cap=cap, tag=args.src, fps=args.fps or 0)
    player.play()
  elif args.command == 'media-server':
    from sciveo.media.pipelines.server import __START_SCIVEO_MEDIA_SERVER__
    __START_SCIVEO_MEDIA_SERVER__()
  elif args.command == 'media-run':
    if args.processor == "audio-plot":
      from sciveo.media.pipelines.processors.audio.audio_extractor_process import plot_audio
      plot_audio(args.width, args.height, args.rate, args.input_path, args.output_path)
  elif args.command == 'watchdog':
    daemons = []
    if args.execute is not None:
      if args.src is None or args.src.startswith("mem"):
        from sciveo.monitoring.watchdog.base import MemoryWatchDogDaemon
        daemons.append(MemoryWatchDogDaemon(threshold_percent=args.threshold, period=args.period, command=args.execute))
      elif args.src.startswith("disk") and args.input_path is not None:
        from sciveo.monitoring.watchdog.base import DiskWatchDogDaemon
        daemons.append(DiskWatchDogDaemon(path=args.input_path, threshold_percent=args.threshold, period=args.period, command=args.execute))
      elif args.src.startswith("process"):
        from sciveo.monitoring.watchdog.process import ProcessWatchDogDaemon
        daemons.append(ProcessWatchDogDaemon(pid=args.pid, process_cmd=args.dst, period=args.period, command=args.execute, thread_inactive_threshold=args.threshold))
    for daemon in daemons:
      daemon.start()
    while(True):
      time.sleep(3600)
  elif args.command == 'predictors-server':
    GlobalConfiguration.set("API_PREDICTORS", None)
    from sciveo.api.server import WebServerDaemon
    daemons = [
      WebServerDaemon(port=args.port)
    ]
    for daemon in daemons:
      debug("starting", type(daemon).__name__)
      daemon.start()
    while(True):
      time.sleep(3600)
  else:
    warning(args.command, "not implemented")

if __name__ == '__main__':
    main()


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import unittest
import numpy as np

from sciveo.tools.logger import *
from sciveo.monitoring.monitor import *


class TestMonitoring(unittest.TestCase):
  def test_cpu(self):
    m = BaseMonitor()
    m.get_cpu_usage()
    print(m.data)

    self.assertTrue("usage per core" in m.data["CPU"])
    self.assertTrue("usage" in m.data["CPU"])

  def test_gpu(self):
    data = {
      "GPU": {
        "raw_lines": [
          "uuid, name, index, power.draw [W], fan.speed [%], memory.total [MiB], memory.used [MiB], memory.free [MiB], utilization.gpu [%], utilization.memory [%], temperature.gpu",
          "GPU-41e0e872-f8d5-95fb-6fd7-8c143c0db7c9, NVIDIA GeForce RTX 3060, 0, 11.57, 0, 12053, 13, 12040, 0, 0, 38",
          "GPU-f7a1d225-58db-e209-aa84-574996d070cd, NVIDIA GeForce RTX 3060, 1, 12.08, 0, 12045, 477, 11568, 0, 39, 36"
        ]
      }
    }
    sample = {}
    config = {
      "CPU": {
        "usage":              {"ratio": 1.0, "metric": "%", "ylim": [0.0, 105.0]},
        "usage per core":     {"ratio": 1.0, "metric": "%", "ylim": [0.0, 105.0]},
      },
      "RAM": {
        "used":               {"ratio": 1.0 / (1024 * 1024 * 1024), "metric": "GB"},
      },
      "GPU": {
        "fan.speed":          {"ratio": 1.0, "metric": "%", "ylim": [0.0, 105.0]},
        "power.draw":         {"ratio": 1.0, "metric": "W", "ylim": [0.0, 250.0]},
        "memory.free":        {"ratio": 1.0 / 1024, "metric": "GB"},
        "memory.used":        {"ratio": 1.0 / 1024, "metric": "GB"},
        "temperature.gpu":    {"ratio": 1.0, "metric": "°C", "ylim": [10, 110]},
        "utilization.gpu":    {"ratio": 1.0, "metric": "%", "ylim": [0.0, 105.0]},
        "utilization.memory": {"ratio": 1.0, "metric": "%", "ylim": [0.0, 105.0]},
      }
    }

    try:
      lines = data["GPU"]["raw_lines"]
      header = lines[0].split(", ")

      keys = []
      gpu_keys = []
      for k in header:
        k_split = k.split(' ')
        key = k_split[0]
        keys.append(key)
        gpu_keys.append(f"GPU {key}")

      for i, k in enumerate(keys):
        if k in config["GPU"]:
          sample[gpu_keys[i]] = []
        else:
          data["GPU"][keys[i]] = []

      for i in range(1, len(lines)):
        line_values = lines[i].split(", ")
        for j, value in enumerate(line_values):
          try:
            value = float(value)
          except ValueError:
            pass
          if keys[j] in config["GPU"]:
            sample[gpu_keys[j]].append(value)
          else:
            data["GPU"][keys[j]].append(value)

      if "memory.total" in data["GPU"]:
        print("memory.total", data["GPU"]["memory.total"])
        for k in ["memory.used", "memory.free"]:
          config["GPU"][k]["ylim"] = [0, max(data["GPU"]["memory.total"]) * config["GPU"][k]["ratio"]]
          print("ylim", config["GPU"][k]["ylim"])
    except Exception as e:
      print("Exception", e)

    print("memory.used", config["GPU"]["memory.used"])
    print("free", config["GPU"]["memory.free"])


from django.http import JsonResponse, HttpResponseForbidden
from django.core.paginator import Paginator
from django.forms.models import model_to_dict
from django.db.models import Q, Case, Value, When, IntegerField
from sciveo.tools.logger import *


class SinglePage_:
  def __init__(self, object_list):
    self.object_list = object_list

class SinglePaginator_:
  def __init__(self, object_list):
    self.object_list = object_list
    self.count = -1
    self.num_pages = -1

  def page(self, page_num):
    return SinglePage_(self.object_list)

def filter_query(request, model_objects, list_query):
  filter_dict = {}
  for q in list_query:
    if q in request.GET:
      q_list = request.GET.getlist(q)
      filter_dict[f"{q}__in"] = q_list
  model_objects = model_objects.filter(**filter_dict)

  query_objects = []
  for q in list_query:
    negative_q = f"~{q}"
    if negative_q in request.GET:
      q_list = request.GET.getlist(negative_q)
      for q_value in q_list:
        query_objects.append(~Q(**{q: q_value}))
  if len(query_objects) > 0:
    query = query_objects.pop()
    for q_obj in query_objects:
      query &= q_obj
    model_objects = model_objects.filter(query)

  return model_objects

def admin_required(view_func):
  def wrapper_func(request, *args, **kwargs):
    if request.user.is_authenticated and request.user.is_superuser:
      return view_func(request, *args, **kwargs)
    else:
      return HttpResponseForbidden()
  return wrapper_func

def data_model_to_dict(object, fields=None, recursive=True):
  data = model_to_dict(object, fields=fields)

  for k in ["created_at", "updated_at"]:
    if hasattr(object, k):
      data[k] = getattr(object, k)

  for k, v in data.items():
    if isinstance(v, list):
      if recursive:
        data[k] = [data_model_to_dict(obj, fields, recursive) for obj in v]
      else:
        data[k] = [str(obj) for obj in v]
  return data

def paginated(request, model_objects, order_default="'id'", already_paginated=False):
  page_current = int(request.GET.get('page', 1))
  page_limit = int(request.GET.get('limit', 10))
  order = request.GET.get('order', order_default)[1:-1]
  fields = request.GET.get('fields', None)
  recursive = "recursive" in request.GET

  try:
    if already_paginated:
      paginator = SinglePaginator_(model_objects)
    else:
      paginator = Paginator(model_objects.order_by(order), page_limit)

    page = paginator.page(page_current)

    data = [data_model_to_dict(obj, fields, recursive) for obj in page.object_list]

    result = {
      "pagination": {
        'page': page_current,
        'limit': page_limit,
        'total_pages': paginator.num_pages,
        'total_count': paginator.count,
      },
      "data": data
    }
  except Exception as e:
    error("paginated", e)
    result = {"error": str(e)}

  return JsonResponse(result)


def filter_on_list(queryset, column_name, filter_list):
  """
  Filters the given queryset based on the values of the specified column
  according to the filter_list.

  Args:
    queryset (QuerySet): The queryset to be sorted.
    column_name (str): The name of the column to sort on.
    filter_list (list): The list defining the desired order.

  Returns:
    QuerySet: The filtered queryset.
  """
  return queryset.filter(Q(**{column_name + '__in': filter_list}))

def sort_on_list(queryset, column_name, sort_list):
  """
  Sorts the given queryset based on the values of the specified column
  according to the order defined in the sort_list.

  Args:
    queryset (QuerySet): The queryset to be sorted.
    column_name (str): The name of the column to sort on.
    sort_list (list): The list defining the desired order.

  Returns:
    QuerySet: The sorted queryset.
  """
  # Create a Case/When expression for each value in the desired order
  # Assign a numerical value to each item in the list
  # Use the Case/When expression to annotate each object with its numerical value
  # Use the annotated value to order the queryset
  return queryset.annotate(
    custom_order=Case(
      *[When(**{column_name: value}, then=Value(index)) for index, value in enumerate(sort_list)],
      default=Value(len(sort_list)),
      output_field=IntegerField()
    )
  ).order_by('custom_order')



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import joblib

import matplotlib.pyplot as plt
import matplotlib.dates as mdates

import numpy as np
import pandas as pd

import tensorflow as tf
from tensorflow import keras

from sciveo.tools.logger import *
from sciveo.tools.array import *
from sciveo.media.ml.encoders.normalizer import *
from sciveo.media.ml.time_series.window_generator import *
from sciveo.media.ml.time_series.dataset import *


class TimeSeriesPredictor:
  def __init__(self, model_path):
    self.model_path = model_path

    self.model_data = joblib.load(self.model_path)

    self.window_generator = WindowGenerator(
      input_width=self.model_data["window"]["input_width"],
      label_width=0,
      shift=self.model_data["window"]["shift"]
    )

  def predict(self, df, x_expand_factor=2):
    normalizer = self.model_data["normalizer"]
    L = self.model_data["window"]["input_width"]
    F = len(self.model_data["columns"])
    predict_df = df[-L:][self.model_data["columns"]]
    predict_df.index = pd.to_datetime(predict_df.index)

    x_plot = list(predict_df.index)
    x_plot = self.expand_dates(x_plot, x_expand_factor)

    predict_df = normalizer.transform(predict_df)

    X = predict_df.values.reshape((1, L, F)).astype('float32')
    X = tf.convert_to_tensor(X)

    predictions = self.model_data["model"](X)
    predictions = normalizer.inverse(predictions)
    X = normalizer.inverse(X)
    debug(f"predict {L} shapes: predict_df[{predict_df.shape}] x_plot[{len(x_plot)}] X[{X.shape}] predictions[{predictions.shape}]")
    return predictions, X, x_plot

  def expand_dates(self, list_dates, expand_factor):
    n = len(list_dates) * expand_factor
    delta = (list_dates[1] - list_dates[0]) # TODO: create more robust delta computation, currently assume equidistant
    expanded_dates = [list_dates[0] + i * delta for i in range(n)]
    return expanded_dates

  def plot(self, predictions, X, x_plot, plot_col_index, labels=None, image_local_path=None, dpi=100, width=640, height=480):
    L = X.shape[1]
    plot_col = self.model_data["columns"][plot_col_index]
    plt.figure(figsize=(12, 8), dpi=dpi)

    plt.subplot(1, 1, 1)
    plt.ylabel(f'{plot_col}')

    plt.plot(
      x_plot[:L], X[0, :, plot_col_index],
      label='X', marker='.', zorder=-10
    )

    plt.scatter(
      x_plot[L:], predictions[0, :, plot_col_index],
      marker='X', edgecolors='k', label='Predictions',
      c='#ff7f0e', s=64
    )

    if labels is not None:
      plt.scatter(
        x_plot[L:], labels[0, :, plot_col_index],
        edgecolors='k', label='Labels', c='#2ca02c', s=64
      )

    plt.xticks(rotation=45)
    plt.tight_layout()
    plt.legend(title=plot_col, loc='upper left')
    plt.xlabel("date time")

    if image_local_path is not None:
      plt.savefig(image_local_path, format='png', dpi=dpi, bbox_inches='tight', pad_inches=0)
      debug("plot savefig", image_local_path)
    else:
      plt.show()
    plt.close()


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import sys
import time
from sciveo.tools.logger import *

def mkdirs(path):
  if not os.path.exists(path):
    os.makedirs(path)
  # else:
  #   debug("mkdirs", path, "exists")

def file_makedirs(file_path):
  return mkdirs(os.path.dirname(file_path))

def touched(path):
  try:
    last_touched = time.time() - os.path.getmtime(path)
  except Exception:
    last_touched = sys.maxsize
  return last_touched

def rm_file(local_path, tag=""):
  try:
    os.remove(local_path)
    debug(tag, "RM", local_path)
  except FileNotFoundError as e:
    exception(e, tag, local_path)
  except Exception as e:
    exception(e, tag, local_path)

def file_name_split(file_path):
  directory, file_name = os.path.split(file_path)
  name, extension = os.path.splitext(file_name)
  return name, extension

def add_suffix_to_filename(file_path, siffix):
  directory, file_name = os.path.split(file_path)
  name, extension = os.path.splitext(file_name)
  new_file_name = f"{name}-{siffix}{extension}"
  new_file_path = os.path.join(directory, new_file_name)
  return new_file_path

def replace_extension_to_filename(file_path, new_ext):
  directory, file_name = os.path.split(file_path)
  name, extension = os.path.splitext(file_name)
  new_file_name = f"{name}.{new_ext}"
  new_file_path = os.path.join(directory, new_file_name)
  return new_file_path

def run_system_cmd(cmd):
  try:
    debug("SYSTEM start [", cmd, "]")
    os.system(cmd)
    debug("SYSTEM finished [", cmd, "]")
  except Exception as e:
    exception(e, "SYSTEM FAILED [", cmd, "]")


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import re

from sciveo.tools.logger import *


class BPETokenizer:
  def __init__(self, max_size):
    self.initial_tokens = 256
    self.max_size = max_size
    self.max_merges = max_size - self.initial_tokens
    self.vocab = {}
    self.merges = {}

  def encode(self, text):
    tokens = list(map(int, text.encode("utf-8")))
    l1 = len(tokens)
    for k, v in self.merges.items():
      self.merge(tokens, k, v)
    debug(f"encoded ratio {len(tokens) / l1:.2}x")
    return tokens

  def decode_token(self, token):
    if token not in self.vocab:
      return [token]

    bigram = self.vocab[token]
    return self.decode_token(bigram[0]) + self.decode_token(bigram[1])

  def decode(self, tokens):
    decoded = []
    for token in tokens:
      decoded += self.decode_token(token)
    return bytes(decoded).decode("utf-8", errors="replace")

  def train(self, text, debug_step=100):
    tokens = list(map(int, text.encode("utf-8")))
    token_id = self.initial_tokens
    debug("max_merges", self.max_merges)
    while(len(self.merges) < self.max_merges):
      current_counts = self.counts(tokens)
      bigram = max(current_counts, key=current_counts.get)
      self.merge(tokens, bigram, token_id)
      self.merges[bigram] = token_id
      self.vocab[token_id] = bigram
      token_id += 1
      if len(self.merges) % debug_step == 0:
        debug("train", f"{len(self.merges)}/{self.max_merges}")

  def fit(self, x):
    return self.train(x)

  def counts(self, tokens):
    result = {}
    for bigram in zip(tokens, tokens[1:]):
      result.setdefault(bigram, 0)
      result[bigram] += 1
    return result

  def merge(self, tokens, bigram, token_id):
    i = 0
    while i < len(tokens) - 1:
      if tokens[i] == bigram[0] and tokens[i + 1] == bigram[1]:
        tokens[i] = token_id
        del tokens[i + 1]
      i += 1



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import base64
import requests

from sciveo.tools.logger import *
from sciveo.tools.configuration import GlobalConfiguration
from sciveo.tools.compress import CompressJsonData


class PredictorRemoteClient:
  def __init__(self, url="http://localhost:8901", verify=True, auth_token=None, api_prefix=None):
    debug("init url", url, verify)
    config = GlobalConfiguration.get()
    if auth_token is None:
      auth_token = config["api_auth_token"]
    if api_prefix is None:
      api_prefix = config["api_prefix"]
    self.auth_token = auth_token
    self.api_prefix = api_prefix
    self.url = f"{url}/{self.api_prefix}/predict"
    self.verify = verify
    requests.packages.urllib3.disable_warnings()

  def predict(self, params):
    try:
      debug("predict", params)
      headers = {"Authorization": f"Bearer {self.auth_token}"}
      response = requests.post(self.url, json=params, headers=headers, verify=self.verify)

      if response.status_code == 200:
        data = response.json()

        if params.get("compressed", 0) > 0 and params["predictor"] in data:
          predicted = data[params["predictor"]]
          predicted = CompressJsonData().decompress(predicted)
          data[params["predictor"]] = predicted
      else:
        error(f"Request [{self.url}] failed with status code {response.status_code}")
        data = {"error": response.status_code}
    except Exception as e:
      exception("predict", e)
      data = {"error": str(e)}
    return data

  def image_encoded(self, image=None, resize_to=(256, 256), local_path=None):
    import cv2
    if local_path is not None:
      image = cv2.imread(local_path)
    if resize_to is not None:
      image = cv2.resize(image, resize_to)
    buffer = cv2.imencode('.jpg', image)[1].tobytes()
    return base64.b64encode(buffer).decode('utf-8')

  def predict_image_embedding(self, image=None, resize_to=(256, 256), local_path=None):
    image_base64 = self.image_encoded(image=image, resize_to=resize_to, local_path=local_path)
    params = {
      'predictor': 'ImageEmbedding', 'compressed': 1,
      'X': [image_base64]
    }
    r = self.predict(params=params)
    return r[params["predictor"]][0]



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import boto3

from sciveo.tools.logger import *
from sciveo.media.pipelines.processors.base import *


class S3MediaDownload(BaseProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)
    self.s3 = boto3.client('s3')
    self.default.update({"output": False, "append-content": False, "update-content": False})

  def process(self, media):
    try:
      # debug("process", media)
      if "bucket" in media:
        remote_path = f"{media['owner']}/{media['key']}"
        local_path = os.path.join(self.base_tmp_path, remote_path.replace("/", "-"))
        if os.path.isfile(local_path):
          debug("SKIP", local_path)
        else:
          debug("process_job content, AWS S3 DWN", media["bucket"], remote_path, local_path)
          self.s3.download_file(media["bucket"], remote_path, local_path)

        self.next_content(media, local_path=local_path)
    except Exception as e:
      exception(e, media)
    return media

  def content_type(self):
    return None

  def name(self):
    return "s3-download"

  def is_append_processor(self):
    return False



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2023
#

import unittest

from sciveo.common.sampling import RandomSampler
from sciveo.common.configuration import Configuration


class TestConfiguration(unittest.TestCase):
  def test_bool(self):
    c = Configuration({
      "bool": True
    })

    self.assertTrue(str(c) == "{\"bool\": true}")



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import json
import gzip
import base64

from sciveo.tools.logger import *


class CompressJsonData:
  def __init__(self) -> None:
    pass

  def compress(self, data):
    json_data = json.dumps(data).encode('utf-8')
    compressed_data = gzip.compress(json_data)
    encoded_data = base64.b64encode(compressed_data).decode('utf-8')
    debug("compress", len(json_data), "->", len(encoded_data))
    return encoded_data

  def decompress(self, encoded_data):
    compressed_data = base64.b64decode(encoded_data)
    decompressed_data = gzip.decompress(compressed_data)
    json_data = decompressed_data.decode('utf-8')
    data = json.loads(json_data)
    debug("decompress", len(encoded_data), "->", len(json_data))
    return data




#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import threading
import uuid
import copy
import json
import datetime

from sciveo.tools.logger import *


class BaseSynchronized:
  def __init__(self, tag=""):
    self.lock_data = threading.Lock()
    if tag:
      tag += "-"
    self._guid = "guid-{}{}-{}".format(tag, datetime.datetime.now().strftime("%Y-%m-%d"), str(uuid.uuid4()).replace("-", ""))

  def guid(self):
    with self.lock_data:
      return self._guid


class ListQueue(BaseSynchronized):
  def __init__(self, tag=""):
    super().__init__(tag)
    self.cv = threading.Condition()
    self.data = []

  def size(self):
    with self.cv:
      return len(self.data)

  def get_data(self):
    with self.cv:
      return copy.deepcopy(self.data)

  def push(self, data):
    with self.cv:
      self.data.append(data)
      self.cv.notify()

  def pop(self, timeout=None):
    with self.cv:
      self.cv.wait_for(predicate=lambda: len(self.data) > 0, timeout=timeout)
      if len(self.data) > 0:
        return self.data.pop(0)
      else:
        raise Exception(f"{type(self).__name__}::POP empty after timeout {timeout}")


class CounterSynchronized(BaseSynchronized):
  def __init__(self, tag=""):
    super().__init__(tag)
    self.data_sync = {}

  def inc(self, k):
    with self.lock_data:
      self.data_sync.setdefault(k, 0)
      self.data_sync[k] += 1

  def dec(self, k):
    with self.lock_data:
      self.data_sync.setdefault(k, 0)
      self.data_sync[k] -= 1

  def data(self):
    with self.lock_data:
      return copy.deepcopy(self.data_sync)


class CallbackSynchronized(BaseSynchronized):
  def __init__(self, tag=""):
    super().__init__(tag)
    self.callbacks = {}

  def hook(self, name, action):
    with self.lock_data:
      self.callbacks.setdefault(name, []).append(action)
      debug(self._guid, "hooked", name, action, self.callbacks)

  def __call__(self, name, data):
    with self.lock_data:
      for callback in self.callbacks.setdefault(name, []):
        try:
          callback(data)
        except Exception as e:
          exception([self], e, name, callback, data)


class DataSynchronized(BaseSynchronized):
  def __init__(self, tag=""):
    super().__init__(tag)
    self.data_sync = {}

  def size(self):
    with self.lock_data:
      return len(self.data_sync)

  def set_one(self, key, value):
    with self.lock_data:
      self.data_sync[key] = value

  def set(self, data):
    with self.lock_data:
      for k, v in data.items():
        self.data_sync[k] = v

  def setdefault(self, data):
    with self.lock_data:
      for k, v in data.items():
        self.data_sync.setdefault(k, v)

  def get(self, key):
    with self.lock_data:
      return self.data_sync.get(key, None)

  def pop(self, key):
    result = None
    with self.lock_data:
      try:
        result = self.data_sync.pop(key)
      except Exception as e:
        result = None
    return result

  def keys(self):
    with self.lock_data:
      return self.data_sync.keys()

  def data(self):
    with self.lock_data:
      return copy.deepcopy(self.data_sync)

  @staticmethod
  def fix_dict(data, list_to_json=[]):
    fixed_data = {}
    for k, v in data.items():
      if type(v) == str:
        v = v.replace("+", " ")
      if k in list_to_json:
        v = v.replace("\'", "\"")
        v = json.loads(v)
      fixed_data[k] = v
    return fixed_data

  @staticmethod
  def data2json(data, keys):
    for key in keys:
      if key in data:
        data[key] = json.loads(data[key].replace("+", " ").replace("\'", "\"").replace("True", "true").replace("False", "false"))



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import time
import sys
from datetime import datetime


class SimpleCounter:
  def __init__(self, count):
    self.counter = 0
    self.current_count = 0
    self.count = count

  def run(self):
    self.counter += 1
    self.current_count += 1
    if self.current_count >= self.count:
      self.current_count = 0

    return self.current_count == 0

  def reset(self):
    self.current_count = 0


class RunCounter(SimpleCounter):
  def __init__(self, count, fn):
    super().__init__(count)
    self.fn = fn

  def run(self):
    if super().run():
      self.fn()


class Timer:
  def __init__(self):
    self.start()

  def start(self):
    self.start_at = time.time()

  def stop(self):
    self.end_at = time.time()
    return self.elapsed()

  def elapsed(self):
    return self.end_at - self.start_at


class TimerDateTime:
  def __init__(self):
    self.start()

  def start(self):
    self.start_at = datetime.now()

  def stop(self):
    self.end_at = datetime.now()
    return self.elapsed()

  def elapsed(self):
    return (self.end_at - self.start_at).total_seconds()


class TimerExpire(Timer):
  def __init__(self, expiry_period):
    super().__init__()
    self.expiry_period = expiry_period

  def finished(self):
    return self.stop() > self.expiry_period


class SimpleStat:
  def __init__(self):
    self.stat = {"min": sys.float_info.max, "max": - sys.float_info.max, "sum": 0, "count": 0, "mean": 0}

  def __dict__(self):
    return self.stat

  def __str__(self):
    return "{}".format(self.stat)

  def __repr__(self):
    return self.__str__()

  def set(self, val):
    if self.stat["min"] > val:
      self.stat["min"] = val
    if self.stat["max"] < val:
      self.stat["max"] = val
    self.stat["sum"] += val
    self.stat["count"] += 1
    self.stat["mean"] = self.stat["sum"] / self.stat["count"]


class ProcessedStatTimer(Timer):
  def __call__(self, num_processed):
    elapsed = self.stop()
    pps = round(num_processed / elapsed, 1)
    elapsed = round(elapsed, 1)
    return elapsed, pps


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2023
#

from setuptools import setup, find_packages
from sciveo.version import __version__

extras_require = {
      'mon': [
        'psutil>=0.0.0',
      ],
      'net': [
        'netifaces>=0.0.0',
        'scapy>=0.0.0',
      ],
      'server': [
        'fastapi>=0.0.0',
        'uvicorn[standard]>=0.0.0',
        'flask>=0.0.0',
        'waitress>=0.0.0',
      ],
      'media': [
        'scikit-learn>=0.0.0', 'scipy>=0.0.0', 'scikit-video>=0.0.0', 'scikit-image>=0.0.0', 'pycryptodome>=0.0.0', 'exifread>=0.0.0', 'qrcode[pil]>=0.0.0',
        'boto3', 'pandas>=0.0.0', 'pika>=0.0.0', 'regex>=0.0.0', 'matplotlib>=0.0.0', 'joblib>=0.0.0', 'tqdm>=0.0.0', 'mss>=0.0.0',
        'ffmpeg-python>=0.0.0', 'opencv-python-headless>=0.0.0', 'opencv-contrib-python-headless>=0.0.0',
      ],
      'media-ml': [
        'tensorflow>=0.0.0', 'keras>=0.0.0',
        'torch>=0.0.0', 'torchvision>=0.0.0',
        'diffusers>=0.0.0', 'transformers>=0.0.0', 'sentence_transformers>=0.0.0', 'accelerate>=0.0.0', 'annoy>=0.0.0',
        'ultralytics>=0.0.0'
      ],
      'web': [
        'django>=0.0.0',
        'djangorestframework>=0.0.0',
        'django-allauth>=0.0.0',
      ],
      'db': [
        'pandas>=0.0.0',
        'psycopg2>=0.0.0',
        'pyarrow>=0.0.0'
      ],
      'power': [
        'pymodbus>=0.0.0',
      ]
}

extras_require['all'] = extras_require['mon'] + extras_require['net'] + extras_require['server']
extras_require['ml'] = extras_require['all'] + extras_require['media'] + extras_require['media-ml']

setup(
    name='sciveo',
    version=__version__,
    packages=find_packages(),
    install_requires=[
      'numpy>=0.0.0',
      'requests>=0.0.0',
    ],
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    extras_require=extras_require,
    py_modules=['sciveo'],
    entry_points={
      'console_scripts': [
        'sciveo=sciveo.cli:main',
      ],
    },
)



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os
import joblib

import numpy as np
import pandas as pd

import tensorflow as tf
from tensorflow import keras

from sciveo.tools.logger import *
from sciveo.tools.array import *
from sciveo.media.ml.encoders.normalizer import *
from sciveo.media.ml.time_series.window_generator import *
from sciveo.media.ml.time_series.dataset import *


class TimeSeriesTrainer:
  def __init__(self, dataset, input_width, label_width, shift):
    self.dataset = dataset

    self.input_width = input_width
    self.label_width = label_width
    self.shift = shift

    self.window_generator = WindowGenerator(
      input_width=self.input_width,
      label_width=self.label_width,
      shift=self.shift
    )

    self.columns = self.dataset.columns
    self.window = {}
    for k, v in self.dataset.dataset.items():
      self.window[k] = self.window_generator.transform(v)

  def build_conv(self):
    num_features = self.dataset.data.shape[1]
    input_layer = tf.keras.layers.Input(shape=(None, num_features))
    x = tf.keras.layers.Lambda(lambda x: x[:, -3:, :])(input_layer)
    x = tf.keras.layers.Conv1D(256, activation='relu', kernel_size=(3))(x)
    x = tf.keras.layers.Dense(self.input_width * num_features, kernel_initializer=tf.initializers.zeros())(x)
    x = tf.keras.layers.Reshape([self.input_width, num_features])(x)

    model = tf.keras.Model(inputs=input_layer, outputs=x)
    debug("model", model.summary())
    return model

  def create(self):
    self.model = self.build_conv()
    self.model.compile(
      loss=tf.keras.losses.MeanSquaredError(),
      optimizer=tf.keras.optimizers.Adam(),
      metrics=[tf.keras.metrics.MeanAbsoluteError()]
    )

  def train(self, max_epochs=20, patience=2, progress_callback=None):
    callbacks=[
        tf.keras.callbacks.EarlyStopping(
          monitor='val_loss',
          patience=patience,
          mode='min'
        )
    ]

    if progress_callback is not None:
      callbacks.append(progress_callback)

    history = self.model.fit(
      self.window["train"],
      validation_data=self.window["val"],
      epochs=max_epochs,
      callbacks=callbacks
    )

    return history

  def save(self, base_model_path, tag):
    model_name = f"model-{self.input_width}-{tag}.timeseries"
    self.model_data = {
      "model": self.model,
      "name": model_name,
      "normalizer": self.dataset.normalizer,
      "columns": self.columns,
      "window": {
          "input_width": self.input_width,
          "label_width": self.label_width,
          "shift": self.shift
      }
    }

    model_path = f"{base_model_path}/{model_name}"
    joblib.dump(self.model_data, model_path)
    debug("saved", model_path)
    return model_name, model_path

  def evaluate(self):
    result = {}
    for k in ["val", "test"]:
      result[k] = self.model.evaluate(self.window["val"], verbose=0)
    return result

  @staticmethod
  def load(dataset, model_path):
    self.model_data = joblib.load(model_path)
    trainer = TimeSeriesTrainer(
      dataset,
      self.model_data["window"]["input_width"],
      self.model_data["window"]["label_width"],
      self.model_data["window"]["shift"]
    )

    trainer.model = self.model_data["model"]
    trainer.normalizer = self.model_data["normalizer"]

    return trainer



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2023
#

import os
import json
from urllib import request, parse
from urllib.error import HTTPError
import requests

from sciveo.tools.logger import *
from sciveo.api.base import *


class APIFileUploader:
  api = None

  def __init__(self, content_type, local_path, parent_guid):
    self.content_type = content_type
    self.local_path = local_path
    self.parent_guid = parent_guid
    if APIFileUploader.api is None:
      APIFileUploader.api = APIRemoteClient()

  def __call__(self):
    file_name = self.local_path.split("/")[-1]
    url_presigned_post = f"aws/s3/presigned_post/?content_type={self.content_type}&file_name={file_name}"
    presigned_post = APIFileUploader.api.GET(url_presigned_post)
    debug("upload presigned_post", url_presigned_post, "=>", presigned_post)
    if "fields" in presigned_post:
      response = None
      with open(self.local_path, 'rb') as fh:
        files = { 'file': (presigned_post['fields']['key'], fh) }
        response = requests.post(presigned_post['url'], data=presigned_post['fields'], files=files)
      if response.status_code == 204:
        debug(self.content_type, self.local_path, "uploaded")
        url_append = f"content/append/?content_type={self.content_type}&name={file_name}&key={presigned_post['fields']['key']}&parent_id={self.parent_guid}"
        result = APIFileUploader.api.GET(url_append)
        debug("upload", url_append, result, "appended")
      else:
        error("upload", self.content_type, self.local_path, "FAIL")




#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2025
#

import cv2
import numpy as np
from datetime import datetime
from sciveo.tools.logger import *


class DrawNone:
  def __call__(self, frame, configuration):
    return frame

class DrawDateTime:
  def __init__(self, position=(10, 10), color=(255, 255, 255), thickness=3, font_scale_factor=0.001, font=cv2.FONT_HERSHEY_SIMPLEX):
    """
    position: tuple (x, y) displacement from top-left
    color: BGR tuple
    thickness: line thickness
    font_scale_factor: relative to frame height (0.0015 is reasonable)
    font: cv2 font
    """
    self.position = position
    self.color = color
    self.thickness = thickness
    self.font = font
    self.font_scale_factor = font_scale_factor

  def __call__(self, frame, configuration):
    """
    Draws date/time (or custom text) on the frame.

    frame: np.ndarray
    text: optional text to draw; if None, current date/time is used
    position, color, thickness: optional overrides
    """
    if frame is None or len(frame.shape) < 2:
      return frame  # invalid frame

    tag = configuration.get("tag", "")

    h, w = frame.shape[:2]

    font_scale = configuration.get("font_scale", max(0.2, h * self.font_scale_factor))
    color = configuration.get("color", self.color)
    thickness = configuration.get("thickness", self.thickness)

    date_text = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    draw_text = f"{tag} {date_text}"

    # Adjust position to avoid going out of frame
    x = min(max(self.position[0], 0), w - 1)
    y = min(max(self.position[1] + int(font_scale * 20), int(font_scale * 20)), h - 1)

    cv2.putText(frame, draw_text, (x, y), self.font, font_scale, color, thickness, cv2.LINE_AA)
    return frame



if __name__ == "__main__":

  cap = cv2.VideoCapture(0)
  drawer = DrawDateTime(tag="test cam", position=(10, 10))

  while True:
    ret, frame = cap.read()
    if not ret:
      break

    frame = drawer(frame)

    cv2.imshow("Frame", frame)
    if cv2.waitKey(1) & 0xFF == ord('q'):
      break

  cap.release()
  cv2.destroyAllWindows()


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import cv2
import numpy as np

from sciveo.tools.common import *
from sciveo.media.pipelines.processors.base import *
from sciveo.media.pipelines.queues import MediaJobState


class VideoAlbum(BaseProcessor):
  def __init__(self, processor_config, max_progress) -> None:
    super().__init__(processor_config, max_progress)
    self.default.update({
      "CRF": 20,
      "image-period": 10,
      "video-frames-count": 10,
      "height": 480, "width": 640,
      "bucket": "smiveo-video",
    })

  def run(self, job, input):
    if len(input) == 0:
      return []

    progress_per_media = self.max_progress / len(input)
    debug("run", job["id"], progress_per_media)

    images = []
    videos = []

    # TODO: Handle different resolutions and aspect ratios
    for i, media in enumerate(input):
      if media["content_type"] == "image":
        images.append(media)
      if media["content_type"] == "video":
        videos.append(media)

    image_period = self["image-period"]
    video_frames_count = self["video-frames-count"]
    height = self["height"]
    width = self["width"]

    parent_guid = job['content_id']
    guid = f"{self.name()}-{parent_guid}"
    key = f"{guid}.mp4"
    bucket = self["bucket"]

    video_local_path = os.path.join(self.base_tmp_path, f"{guid}.mp4")
    video = VideoWriterFFMPEG.new(video_local_path, crf=str(self["CRF"]))

    for media in images:
      try:
        frame = cv2.imread(media["local_path"])
        frame_resized = cv2.resize(frame, (width, height), interpolation=cv2.INTER_AREA)
        for i in range(image_period):
          video.write(frame_resized)
      except Exception as e:
        exception(e)
      MediaJobState.queue().inc_progress(job["id"], progress_per_media)

    for media in videos:
      try:
        cap = cv2.VideoCapture(media["local_path"])
        frames_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
        frames_idx = np.linspace(0, frames_count - 1, video_frames_count, dtype=int)

        for frame_idx in frames_idx:
          cap.set(cv2.CAP_PROP_POS_FRAMES, frame_idx)
          ret, frame = cap.read()
          if not ret:
            break

          frame_resized = cv2.resize(frame, (width, height), interpolation=cv2.INTER_AREA)
          video.write(frame_resized)

        cap.release()
      except Exception as e:
        exception(e)
      MediaJobState.queue().inc_progress(job["id"], progress_per_media)

    video.close()

    media = {
      "guid": guid,
      "parent": parent_guid,
      "content_type": "video",
      "owner": job["owner"],
      "local_path": video_local_path,
      "w": width, "h": height,
      "height": height,
      "key": key,
      "bucket": bucket,
      "processor": self.name(),
      "layout": {"name": self.name(), "height": height, **self["layout"]}
    }

    if self["output"]:
      job["output"].append(media)
    if self["append-content"]:
      job["append-content"].append(media)
    return [media]

  def content_type(self):
    return ["video", "image"]

  def name(self):
    return "video-album"

  def is_append_processor(self):
    return False


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import os

from sciveo.tools.logger import *
from sciveo.tools.http import *


class BaseLayout:
  def post(self, url, data):
    headers = {
      "Auth-Token": os.environ['MEDIA_API_AUTH_TOKEN']
    }
    response = POST(url, data=data, headers=headers)
    debug("post response", response)
    return response


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2024
#

import numpy as np
from scipy.io import wavfile
from scipy.signal import butter, lfilter
import matplotlib.pyplot as plt

from sciveo.tools.logger import *
from sciveo.tools.common import *


class AudioFFT:
  def decimate(self, audio_data, ratio):
    # Define the cutoff frequency and the order of the filter
    cutoff = 0.1
    order = 4
    # Create the Butterworth filter
    b, a = butter(order, cutoff, btype='low', analog=False, output='ba')
    # Apply the filter to the signal
    filtered_signal = lfilter(b, a, audio_data)
    # Decimate the filtered signal
    return filtered_signal[::ratio]

  def fft(self, sample_rate, audio_data):
    debug("audio_data.shape", audio_data.shape)
    # Check for stereo, pick the first channel.
    if len(audio_data.shape) > 1 and audio_data.shape[1] > 1:
      audio_data = audio_data[:,0]

    if audio_data.shape[0] > 3000000:
      #audio_data = self.decimate(audio_data, 100)
      audio_data = audio_data[:3000000]

    debug("audio_data.shape", audio_data.shape)

    result = {}
    result["length"] = audio_data.shape[0] / sample_rate

    fft = np.fft.fft(audio_data)
    fft_mag = (fft.real**2 + fft.imag**2) ** 0.5

    result["audio_time"] = np.linspace(0., result["length"], audio_data.shape[0])
    freq = np.fft.fftfreq(result["audio_time"].shape[-1]) * sample_rate

    result["audio_data"] = self.normalize(audio_data)

    result["fft"], result["freq"] = self.half_fft(fft_mag, freq)
    result["fft"] = self.normalize(result["fft"])

    result["fft_bins"], result["freq_bins"] = self.bin_fft(result["fft"], result["freq"], bins=16)
    result["fft_bins"] = self.normalize(result["fft_bins"])

    result["hist"], result["hist_bins"] = np.histogram(np.absolute(result["audio_data"]), bins=4)
    result["hist"] = self.normalize(result["hist"])

    result["freq_fft_max"] = result["freq"][np.argmax(result["fft"])] * 2 / sample_rate
    result["freq_fft_min"] = result["freq"][np.argmin(result["fft"])] * 2 / sample_rate

    return result

  def half_fft(self, fft, freq, cutoff_freq=0):
    bool_freq = freq >= cutoff_freq
    return fft[bool_freq], freq[bool_freq]

  def bin_fft(self, fft, freq, bins=16):
    l = len(fft)
    l1 = int(l / bins)
    l2 = l1 * bins
    fft = fft[l - l2:].reshape((bins, l1)).sum(axis=1)
    freq = freq[l - l2:].reshape((bins, l1)).max(axis=1)
    return fft, freq

  def binning(self, data, bins=16):
    l = len(data)
    l1 = int(l / bins)
    l2 = l1 * bins
    data = data[l - l2:].reshape((bins, l1)).sum(axis=1)
    return data

  def normalize(self, data):
    # return data / np.linalg.norm(data)
    # return data / max(data) # scaling here...
    min_data = data.min()
    max_data = data.max()

    if max_data - min_data < 1:
      return data

    return (data - min_data) / (max_data - min_data)


def plot_audio(width, height, sampling_rate, aac_audio_local_path, image_audio_local_path):
  debug("Audio Extractor", aac_audio_local_path)
  try:
    wav_audio_local_path = aac_audio_local_path.replace(".aac", ".wav")
    cmd = f"ffmpeg -i '{aac_audio_local_path}' -ar {sampling_rate} '{wav_audio_local_path}'"
    os.system(cmd)

    sample_rate, audio_data = wavfile.read(wav_audio_local_path)
    debug("Audio Extractor sample_rate", sample_rate, "audio_data shape", audio_data.shape)
    os.remove(wav_audio_local_path)

    # Check for stereo, pick the first channel.
    if len(audio_data.shape) > 1 and audio_data.shape[1] > 1:
      audio_data = audio_data[:,0]

    dpi = 100
    fig = plt.figure(figsize=(int(width)/dpi, int(height)/dpi), dpi=dpi)
    ax1 = fig.add_subplot(2, 1, 1)
    ax2 = fig.add_subplot(2, 1, 2)

    ax1.plot(audio_data)
    ax1.set_xlabel('samples')
    ax1.set_ylabel('Amplitude')
    ax1.set_title("Audio Time domain")

    fft = AudioFFT().fft(sample_rate, audio_data)

    ax2.plot(fft["freq"], fft["fft"])
    ax2.set_xlabel('frequency')
    ax2.set_ylabel('Amplitude')
    ax2.set_title("Audio Frequency domain")

    plt.subplots_adjust(hspace=0.5)
    plt.savefig(image_audio_local_path, dpi=dpi, bbox_inches='tight', pad_inches=0)
  except Exception as e:
    exception(e, aac_audio_local_path)



#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2025
#


import os
import hmac
import base64
import struct
import hashlib
import time

from sciveo.tools.logger import *


class AuthTOTP:
  def __init__(self, secret: str, interval: int = 30, digits: int = 6, algo=hashlib.sha1):
    """
    secret: the private key
    interval: time step in seconds (default 30)
    digits: number of digits in code (default 6)
    algo: hash function (default SHA1, per RFC 6238)
    """
    self.secret = secret
    self.interval = interval
    self.digits = digits
    self.algo = algo

  @staticmethod
  def new_private_key(length: int = 20) -> str:
    """
    Generate a new Base32 secret key.
    """
    raw = os.urandom(length)
    return base64.b32encode(raw).decode("utf-8").replace("=", "")

  def _generate_otp(self, for_time: int) -> int:
    """
    Internal: generate OTP for given time counter.
    """
    pad_len = (8 - len(self.secret) % 8) % 8
    secret_padded = self.secret + ("=" * pad_len)
    key = base64.b32decode(secret_padded, casefold=True)

    counter = struct.pack(">Q", for_time)
    hmac_digest = hmac.new(key, counter, self.algo).digest()

    offset = hmac_digest[-1] & 0x0F
    code = struct.unpack(">I", hmac_digest[offset:offset+4])[0] & 0x7FFFFFFF
    return code % (10 ** self.digits)

  def next_token(self, for_time: int = None) -> str:
    """
    Return the TOTP token for current (or given) time.
    """
    if for_time is None:
      for_time = int(time.time())
    counter = for_time // self.interval
    return str(self._generate_otp(counter)).zfill(self.digits)

  def verify_token(self, token: str, for_time: int = None, window: int = 1) -> bool:
    """
    Verify a token, allowing ±window intervals (to handle clock skew).
    """
    if for_time is None:
      for_time = int(time.time())
    counter = for_time // self.interval
    for delta in range(-window, window + 1):
      candidate = str(self._generate_otp(counter + delta)).zfill(self.digits)
      # debug("verify", token, "delta", delta, "candidate", candidate)
      if hmac.compare_digest(candidate, token):
        return True
    return False



import math
import cv2


PI = 3.141592653589793

def to_radians(angle):
  return angle * PI / 180

def line_perpendicular(p1, p2, l):
  x1, y1 = p1
  x2, y2 = p2

  x3 = x2 - x1
  y3 = y2 - y1

  mag = math.sqrt(x3 * x3 + y3 * y3)
  x3 /= mag
  y3 /= mag

  x4 = x2 - y3 * l
  y4 = y2 + x3 * l
  x5 = x2 + y3 * l
  y5 = y2 - x3 * l
  return (int(x4), int(y4)), (int(x5), int(y5))

def frame_bounds_rewind(w, h, x, y):
  if x > w:
    x = 0
  if x < 0:
    x = w
  if y > h:
    y = 0
  if y < 0:
    y = h
  return x, y


#
# Pavlin Georgiev, Softel Labs
#
# This is a proprietary file and may not be copied,
# distributed, or modified without express permission
# from the owner. For licensing inquiries, please
# contact pavlin@softel.bg.
#
# 2023
#

import threading
import signal
import time

from sciveo.tools.logger import *
from sciveo.tools.synchronized import ListQueue
from sciveo.api.upload import APIFileUploader


class SignalHandler:
  def __init__(self, daemons):
    self.daemons = daemons

    signal.signal(signal.SIGINT, self.handler_int)
    signal.signal(signal.SIGUSR1, self.handler_usr1)

  def handler_int(self, signum, frame):
    info("INT signal received", signum, frame)
    for daemon in self.daemons:
      daemon.stop()

  def handler_usr1(self, signum, frame):
    info("USR1 signal received", signum, frame)


class DaemonBase:
  def __init__(self, num_threads=1, period=0):
    self.is_running = False
    self.is_started = False
    self.num_threads = num_threads
    self.period = period

  def start(self):
    if self.is_started:
      return

    self.is_running = True
    self.is_started = True

    self.threads = []
    for i in range(self.num_threads):
      T = threading.Thread(target = self.safe_run)
      T.setDaemon(True)
      T.start()
      self.threads.append(T)

  def stop(self):
    self.is_running = False

  def finalise(self):
    pass

  def join(self):
    for i in range(self.num_threads):
      self.threads[i].join()

  def loop(self):
    pass

  def run(self):
    while(self.is_running):
      try:
        self.loop()
      except Exception as e:
        exception(e)
      time.sleep(self.period)

  def safe_run(self):
    try:
      self.run()
    except Exception as e:
      exception(e)


class TasksDaemon(DaemonBase):
  current = None
  queue = ListQueue("tasks")

  def loop(self):
    task = TasksDaemon.queue.pop()
    task()


def __upload_content__(content_type, local_path, parent_guid):
  TasksDaemon.queue.push(APIFileUploader(content_type, local_path, parent_guid))